#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scripts for making specific models.
"""

import warnings
import toml
from datetime import datetime
import tensorflow as tf
import tensorflow.keras as ks
import tensorflow.keras.layers as layers

from orcanet.builder_util.builders import BlockBuilder


class ModelBuilder:
    """
    Build and compile a keras model from a toml file, using OrcaNet building blocks.

    The input of the model can match the dimensions of the input
    data given to the Organizer taking into account the sample
    modifier.

    Attributes
    ----------
    configs : list
        List with keywords for building each layer block in the model.
    defaults : dict
        Default values for the layer blocks in the model.
    optimizer : str or Optimizer
        Optimizer for training the model. Can be a string like "adam" (or
        "keras:adam" for the default keras variant), or an object derived
        from ks.optimizers.Optimizer.
    compile_opt : dict
        Keys: Names of the output layers of the model.
        Values: Loss function, optionally weight and optionally metric of
        each output layer.
        Format: { layer_name : { loss_function:, weight:, metrics: } }
        The loss_function is a string or a function, the weight is a float
        and metrics is a list of functions/strings.
    optimizer_args : dict, optional
        Kwargs for the optimizer. Not used when an optimizer object is given.
    input_opts : dict
        Specify options for the input of the model.

    Methods
    -------
    build
        Build the network using an instance of Organizer.
    build_with_input
        Build the network without an Organizer, just using given input shapes.
    compile
        Compile a model with the optimizer settings given in the model_file.

    """
    def __init__(self, model_file, **custom_blocks):
        """
        Read out parameters for creating models with OrcaNet from a toml file.

        Parameters
        ----------
        model_file : str
            Path to the model toml file.
        custom_blocks
            For building models with custom blocks in the toml:
            Custom block names as kwargs ('toml name'='block').

        """
        file_content = toml.load(model_file)
        self.custom_blocks = custom_blocks

        try:
            if "model" in file_content:
                model_args = file_content["model"]
                self.configs = model_args.pop('blocks')
                self.input_opts = model_args.pop("input_opts", {})
                self.defaults = model_args

            elif "body" in file_content:
                # legacy
                self._compat_init(file_content)

            self.optimizer = None
            self.compile_opt = None
            self.optimizer_args = {}
            if "compile" in file_content:
                compile_sect = file_content["compile"]
                self.optimizer = compile_sect.pop("optimizer", None)
                self.compile_opt = compile_sect.pop("losses", None)
                self.optimizer_args = compile_sect

        except KeyError as e:
            if len(e.args) == 1:
                option = e.args[0]
            else:
                option = e.args
            raise KeyError("Missing parameter in toml model file: "
                           + str(option)) from None

    def _compat_init(self, file_content):
        warnings.warn(
            "The format of this model toml file is deprecated, consider "
            "updating it to the new format (see online docu)."
        )
        # legacy
        body = file_content["body"]
        if "architecture" in body:
            arch = body.pop("architecture")
            if arch != "single":
                raise ValueError("architecture keyword is deprecated")
        self.configs = body.pop('blocks')
        self.defaults = body

        if "head" in file_content:
            head = file_content["head"]
            head_arch = head.pop("architecture")
            head_arch_args = head.pop("architecture_args")
            head_args = head

            head_block_config = head_arch_args
            head_block_config["type"] = head_arch
            self.configs.append({**head_block_config, **head_args})

    def build(self, orga, log_comp_opts=False, verbose=False):
        """
        Build the network using an instance of Organizer.

        Input layers will be adapted to the input files in the organizer.
        Can also add the matching modifiers and custom objects to the orga.

        Parameters
        ----------
        orga : orcanet.core.Organizer
            Contains all the configurable options in the OrcaNet scripts.
        log_comp_opts : bool
            If the info used for the compilation of the model should be
            logged to the log.txt.
        verbose : bool
            Print info about the building process?

        Returns
        -------
        model : keras model
            The network.

        """
        if orga.cfg.fixed_batchsize:
            if "batchsize" in self.input_opts and \
                    self.input_opts["batchsize"] != orga.cfg.batchsize:
                raise ValueError(
                    f"Batchsize in input_opts is {self.input_opts['batchsize']}, "
                    f"but in cfg its {orga.cfg.batchsize}")
            self.input_opts["batchsize"] = orga.cfg.batchsize

        with orga.get_strategy().scope():
            model = self.build_with_input(
                orga.io.get_input_shapes(),
                compile_model=True,
                custom_objects=orga.cfg.get_custom_objects(),
                verbose=verbose,
            )

        if log_comp_opts:
            self.log_model_properties(orga)
        model.summary()
        return model

    def build_with_input(self, input_shapes,
                         compile_model=True,
                         custom_objects=None,
                         verbose=False):
        """
        Build the network with given input shapes.

        Parameters
        ----------
        input_shapes : dict
            Keys: Name of the inputs of the model.
            Values: Their shape without the batchsize.
        compile_model : bool
            Compile the model?
        custom_objects : dict, optional
            Custom objects to use during compiling.
        verbose : bool
            Print info about the building process?

        Returns
        -------
        model : ks.Model
            The network.

        """
        builder = BlockBuilder(
            self.defaults, verbose=verbose, input_opts=self.input_opts,
            **self.custom_blocks)
        model = builder.build(input_shapes, self.configs)

        if compile_model:
            self.compile_model(model, custom_objects=custom_objects)

        return model

    # def merge_models(self, model_list, trainable=False, stateful=True,
    #                  no_drop=True):
    #     """
    #     Concatenate two or more single input cnns to a big one.
    #
    #     It will explicitly look for a Flatten layer and cut after it,
    #     Concatenate all models, and then add the head layers.
    #
    #     Parameters
    #     ----------
    #     model_list : list
    #         List of keras models to stitch together.
    #     trainable : bool
    #         Whether the layers of the loaded models will be trainable.
    #     stateful : bool
    #         Whether the batchnorms of the loaded models will be stateful.
    #     no_drop : bool
    #         If true, rate of dropout layers from loaded models will
    #         be set to zero.
    #
    #     Returns
    #     -------
    #     model : keras model
    #         The uncompiled merged keras model.
    #
    #     """
    #     # Get the input and Flatten layers in each of the given models
    #     input_layers, flattens = [], []
    #     for i, model in enumerate(model_list):
    #         if len(model.inputs) != 1:
    #             raise ValueError(
    #                 "model input is not length 1 {}".format(model.inputs))
    #         input_layers.append(model.input)
    #         flatten_found = 0
    #         for layer in model.layers:
    #             layer.trainable = trainable
    #             layer.name = layer.name + '_net_' + str(i)
    #             if isinstance(layer, layers.BatchNormalization):
    #                 layer.stateful = stateful
    #             elif isinstance(layer, layers.Flatten):
    #                 flattens.append(layer.output)
    #                 flatten_found += 1
    #         if flatten_found != 1:
    #             raise TypeError(
    #                 "Expected 1 Flatten layer but got " + str(flatten_found))
    #
    #     # attach new head
    #     x = layers.Concatenate()(flattens)
    #     builder = BlockBuilder(body_defaults=None,
    #                            head_defaults=self.head_args)
    #     output_layer = builder.attach_output_layers(x, self.head_arch,
    #                                                 flatten=False,
    #                                                 **self.head_arch_args)
    #
    #     model = ks.models.Model(input_layers, output_layer)
    #     if no_drop:
    #         model = change_dropout_rate(model, before_concat=0.)
    #
    #     return model

    def compile_model(self, model, custom_objects=None):
        """
        Compile a model with the optimizer settings given as the attributes.

        Parameters
        ----------
        model : ks.model
            A keras model.
        custom_objects : dict or None
            Maps names (strings) to custom loss functions.

        Returns
        -------
        model : keras model
            The compiled (or recompiled) keras model.

        """
        if any((self.optimizer is None, self.compile_opt is None)):
            raise ValueError("Can not compile, need optimizer name and losses")

        loss_functions, loss_weights, loss_metrics = {}, {}, {}
        for layer_name, layer_info in self.compile_opt.items():
            # Replace the str function name with actual function if it is custom
            loss_function = layer_info['function']
            if custom_objects is not None and loss_function in custom_objects:
                loss_function = custom_objects[loss_function]
            loss_functions[layer_name] = loss_function

            # Use given weight, else use default weight of 1
            if 'weight' in layer_info:
                weight = layer_info['weight']
            else:
                weight = 1.0
            loss_weights[layer_name] = weight

            # Use given metrics, else use no metrics
            if 'metrics' in layer_info:
                metrics = layer_info['metrics']
            else:
                metrics = []
            if custom_objects is not None:
                for i, metric in enumerate(metrics):
                    if metric in custom_objects:
                        metrics[i] = custom_objects[metric]

            loss_metrics[layer_name] = metrics

        optimizer = self._get_optimizer()
        model.compile(loss=loss_functions, optimizer=optimizer,
                      metrics=loss_metrics, loss_weights=loss_weights)
        return model

    def log_model_properties(self, orga):
        """
        Writes the compile_opt config to the full log file.
        """
        lines = list()
        lines.append('-' * 60)
        time = datetime.now().strftime('%Y-%m-%d  %H:%M:%S')
        lines.append('-' * 19 + " {} ".format(time) + '-' * 19)
        lines.append("A model has been built using the model builder with the following configurations:\n")
        lines.append('Loss functions: ')
        for key in self.compile_opt:
            lines.append(key + ': ' + str(self.compile_opt[key]))
        lines.append('\n')
        orga.io.print_log(lines)

    def _get_optimizer(self):
        if not isinstance(self.optimizer, str):
            if self.optimizer_args:
                warnings.warn(
                    "Custom callback used, optimizer_args are ignored: " +
                    str(self.optimizer_args)
                )
            return self.optimizer
        if self.optimizer == 'adam':
            optimizer = get_adam(**self.optimizer_args)
        elif self.optimizer == 'sgd':
            optimizer = get_sgd(**self.optimizer_args)
        elif self.optimizer.startswith("keras:"):
            optimizer = getattr(
                ks.optimizers, self.optimizer.split("keras:")[-1]
            )(**self.optimizer_args)
        else:
            raise NameError('Unknown optimizer name ({})'.format(self.optimizer))
        return optimizer


def get_adam(beta_1=0.9, beta_2=0.999, epsilon=0.1, decay=0.0, **kwargs):
    # epsilon=1 for deep networks
    return ks.optimizers.Adam(beta_1=beta_1,
                              beta_2=beta_2,
                              epsilon=epsilon,
                              decay=decay,
                              **kwargs)


def get_sgd(momentum=0.9, decay=0, nesterov=True, **kwargs):
    return ks.optimizers.SGD(momentum=momentum,
                             decay=decay,
                             nesterov=nesterov,
                             **kwargs)


def _change_dropout_rate(model, before_concat, after_concat=None):
    """
    Change the dropout rate in a model.

    # TODO untested for tf 2.x!

    Only for models with a concatenate layer, aka multiple
    single input models that were merged together.

    Parameters
    ----------
    model : keras model

    before_concat : float
        New dropout rate before the concatenate layer in the model.
    after_concat : float or None
        New dropout rate after the concatenate layer. None will leave the
        dropout rate there as it was.

    """
    ch_bef, ch_aft, concat_found = 0, 0, 0

    for layer in model.layers:
        if isinstance(layer, layers.Dropout):
            if concat_found == 0:
                layer.rate = before_concat
                ch_bef += 1
            else:
                layer.rate = after_concat
                ch_aft += 1

        elif isinstance(layer, layers.Concatenate):
            concat_found += 1
            if after_concat is None:
                break

    if concat_found != 1:
        raise TypeError("Expected 1 Flatten layer but got " + str(concat_found))
    clone = ks.models.clone_model(model)
    clone.set_weights(model.get_weights())
    print("Changed dropout rates of {} layers before and {} layers after "
          "Concatenate.".format(ch_bef, ch_aft))
    return clone



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from unittest import TestCase
from unittest.mock import MagicMock, patch
import tensorflow.keras as ks
import tensorflow.keras.layers as layers

from orcanet.core import Organizer
from orcanet.model_builder import ModelBuilder
from orcanet_contrib.custom_objects import get_custom_objects


class TestModel(TestCase):
    def setUp(self):
        """
        Make a .temp directory in the current working directory, generate
        dummy data in it and set up the cfg object.

        """
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.model_file = os.path.join(self.data_folder, "model_builder_test_CNN_cat.toml")

        self.input_shapes = {
            "input_A": (3, 3, 3, 3),
        }

        orga = Organizer(".")
        orga.io.get_input_shapes = MagicMock(return_value=self.input_shapes)
        orga.cfg.custom_objects = get_custom_objects()

        self.orga = orga

    def test_model_setup_CNN_model(self):
        orga = self.orga

        builder = ModelBuilder(self.model_file)
        model = builder.build(orga)

        self.assertEqual(model.input_shape[1:], self.input_shapes["input_A"])
        self.assertEqual(model.output_shape[1:], (2, ))
        self.assertEqual(len(model.layers), 14)
        self.assertEqual(model.optimizer.epsilon, 0.2)

    def test_model_setup_CNN_model_custom_callback(self):
        builder = ModelBuilder(self.model_file)
        builder.optimizer = ks.optimizers.SGD()
        model = builder.build(self.orga)
        self.assertIsInstance(model.optimizer, ks.optimizers.SGD)

    @patch('orcanet.model_builder.toml.load')
    def test_load_optimizer(self, mock_toml_load):
        def toml_load(file):
            return file
        mock_toml_load.side_effect = toml_load

        file_cntn = {
            "model": {"blocks": None},
            "compile": {"optimizer": "keras:Adam", "losses": None, "lr": 1.0}
        }
        builder = ModelBuilder(file_cntn)
        opti = builder._get_optimizer()

        target = {
            'name': 'Adam',
            'learning_rate': 1.0,
            'beta_1': 0.8999999761581421,
            'beta_2': 0.9990000128746033,
            'decay': 0.0,
            'epsilon': 1e-07,
            'amsgrad': False
        }

        for k, v in opti.get_config().items():
            self.assertAlmostEqual(v, target[k])

    @patch('orcanet.model_builder.toml.load')
    def test_custom_blocks(self, mock_toml_load):
        def toml_load(file):
            return file
        mock_toml_load.side_effect = toml_load

        file_cntn = {
            "model": {"blocks": [{"type": "my_custom_block", "units": 10}, ]},
        }
        builder = ModelBuilder(file_cntn, my_custom_block=layers.Dense)
        model = builder.build_with_input({"a": (10, 1)}, compile_model=False)

        self.assertIsInstance(model.layers[-1], layers.Dense)
        self.assertEqual(model.layers[-1].get_config()["units"], 10)

    # def test_merge_models(self):
    #     def build_model(inp_layer_name, inp_shape):
    #         inp = layers.Input(inp_shape, name=inp_layer_name)
    #         x = layers.Convolution3D(3, 3)(inp)
    #         x = layers.Flatten()(x)
    #         out = layers.Dense(1, name="out_0")(x)
    #
    #         model = Model(inp, out)
    #         return model
    #
    #     model_file = os.path.join(self.data_folder, self.model_file)
    #     builder = ModelBuilder(model_file)
    #     model1 = build_model("inp_A", self.input_shapes["input_A"])
    #     model2 = build_model("inp_B", self.input_shapes["input_A"])
    #     merged_model = builder.merge_models([model1, model2])
    #
    #     for layer in model1.layers + model2.layers:
    #         if isinstance(layer, layers.Dense):
    #             continue
    #         merged_layer = merged_model.get_layer(layer.name)
    #         for i in range(len(layer.get_weights())):
    #             self.assertTrue(np.array_equal(layer.get_weights()[i],
    #                                            merged_layer.get_weights()[i]))



import orcanet.misc as misc

# for loading via toml
dmods, register = misc.get_register()


@register
def as_array(info_blob):
    """
    Save network output as ndarrays to h5. This is the default dataset modifier.

    Every output layer will get one dataset each for both the label and
    the prediction. E.g. if the model has an output layer called "energy",
    the datasets "label_energy" and "pred_energy" will be made.

    """
    datasets = dict()

    y_pred = info_blob["y_pred"]
    for out_layer_name in y_pred:
        datasets["pred_" + out_layer_name] = y_pred[out_layer_name]

    ys = info_blob.get("ys")
    if ys is not None:
        for out_layer_name in ys:
            datasets["label_" + out_layer_name] = ys[out_layer_name]

    y_values = info_blob.get("y_values")
    if y_values is not None:
        datasets['y_values'] = y_values

    return datasets


@register
def as_recarray(info_blob):
    """
    Save network output as recarray to h5. Intended for when network
    outputs are 2D, i.e. (batchsize, X).

    Output from network:
    Dict with arrays, shapes (batchsize, x_i).
    E.g. {"foo": ndarray, "bar": ndarray}

    dtypes that will get saved to h5:
    (foo_1, foo_2, ..., bar_1, bar_2, ... )

    """
    datasets = dict()
    datasets["pred"] = misc.dict_to_recarray(info_blob.get("y_pred"))

    ys = info_blob.get("ys")
    if ys is not None:
        datasets["true"] = misc.dict_to_recarray(ys)

    y_values = info_blob.get("y_values")
    if y_values is not None:
        datasets['y_values'] = y_values  # is already a structured array

    return datasets


@register
def as_recarray_dist(info_blob):
    """
    Save network output as recarray to h5. Intended for when network
    outputs are distributions and thus 3D (for example when using
    OutputRegNormal as output layer block).
    I.e. (batchsize, 2, X), with [:, 0] being mu and [:, 1] being std.

    Example output from network:
    shape {"A": (bs, 2), "B": (bs, 2, 3)}
        [:, 0] is reco, [:, 1] is err

    dtypes that will get saved to h5:
    A_1, A_err_1, B_1, B_2, B_3, B_err_1, B_err_2, B_err_3

    """
    y_pred = info_blob["y_pred"]
    datas = {}
    for output_name, array in y_pred.items():
        # [:, 0] is mu and [:, 1] is err
        datas[output_name] = array[:, 0]
        datas[f"{output_name}_err"] = array[:, 1]
    info_blob["y_pred"] = datas

    ys = info_blob.get("ys")
    if ys is not None:
        # errs for the trues are just padded, so skip
        datas = {}
        for output_name, array in ys.items():
            datas[output_name] = array[:, 0]
        info_blob["ys"] = datas

    return as_recarray(info_blob)


@register
def as_recarray_dist_split(info_blob):
    """
    Save network output as recarray to h5. Intended for networks that
    output recos and errs in seperate towers (for example when using
    OutputRegNormalSplit as output layer block).

    Example output from network:
    shape {"A": (bs, 1), "A_err": (bs, 2, 1),
           "B": (bs, 3), "B_err": (bs, 2, 3)}
    In "A_err": [:, 0] is mu, [:, 1] is sigma

    dtypes that will get saved to h5:
    A_1, A_err_1, B_1, B_1_err, B_2, B_err_2, ...

    """
    def transform(network_output):
        """ Skip A and rename A_err to A. """
        transformed = {}
        for output_name, output_value in network_output.items():
            if output_name.endswith("_err"):
                transformed[output_name[:-4]] = output_value
        return transformed

    info_blob["y_pred"] = transform(info_blob["y_pred"])
    if info_blob.get("ys") is not None:
        info_blob["ys"] = transform(info_blob["ys"])

    return as_recarray_dist(info_blob)



""" Odds and ends. """
import os
import inspect
import numpy as np


def get_register():
    """ E.g. for storing orcanet layer blocks as custom objects. """
    saved = {}

    def register(obj):
        saved[obj.__name__] = obj
        return obj
    return saved, register


def from_register(toml_entry, register):
    """
    Get an initilized object via a toml entry.
    Used for loading orcanet built-in sample modifiers etc.

    Parameters
    ----------
    toml_entry : str or dict or list
        The 'sample_modifier' given in the config toml.
        E.g., to initialize "obj_name" from register, these are possible formats:
        "obj_name"
        ["obj_name", True]
        ["obj_name", {"setting_1": True}]
        {"name": "obj_name", "setting_1": True}
    register : dict
        Maps class names to class references.

    """
    args, kwargs = [], {}
    if isinstance(toml_entry, str):
        name = toml_entry
    elif isinstance(toml_entry, dict):
        if "name" not in toml_entry:
            raise KeyError(f"missing entry in dict: 'name', given: {toml_entry}")
        name = toml_entry["name"]
        kwargs = {k: v for k, v in toml_entry.items() if k != "name"}
    else:
        name = toml_entry[0]
        if len(toml_entry) == 2 and isinstance(toml_entry[1], dict):
            kwargs = toml_entry[1]
        else:
            args = toml_entry[1:]

    obj = register[name]
    try:
        if inspect.isfunction(obj):
            if args or kwargs:
                raise TypeError(f"Can not pass arguments to function ({args}, {kwargs})")
            return obj
        else:
            return obj(*args, **kwargs)
    except TypeError:
        raise TypeError(f"Error initializing {obj}")


def dict_to_recarray(array_dict):
    """
    Convert a dict with np arrays to a 2d recarray.
    Column names are derived from the dict keys.

    Parameters
    ----------
    array_dict : dict
        Keys: string
        Values: ND arrays, same length and number of dimensions.
            All dimensions expect first will get flattened.

    Returns
    -------
    The recarray.

    """
    column_names, arrays = [], []
    for key, array in array_dict.items():
        if len(array.shape) == 1:
            array = np.expand_dims(array, -1)
        elif len(array.shape) > 2:
            array = np.reshape(array, (len(array), -1))
        for i in range(array.shape[-1]):
            arrays.append(array[:, i])
            column_names.append(f"{key}_{i+1}")
    return np.core.records.fromarrays(arrays, names=column_names)


def to_ndarray(x, dtype="float32"):
    """ Turn recarray to ndarray. """
    new_dtype = [(name, dtype) for name in x.dtype.names]
    new_shape = (len(x), len(x.dtype.names))
    return np.ascontiguousarray(x).astype(new_dtype).view(dtype).reshape(new_shape)


def find_file(directory, filename):
    """ Look for file in given directoy. Error if there are multiple. """
    found = []
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file == filename:
                found.append(os.path.join(root, file))
    if len(found) >= 2:
        raise ValueError(
            f"Can not find {filename}: More than one file found ({found})")
    elif len(found) == 0:
        return None
    else:
        fpath = found[0]
        print(f"Found {fpath}")
        return fpath



import tensorflow as tf
from tensorflow.keras.models import Model
import tensorflow.keras.layers as layers

from orcanet.builder_util.builders import BlockBuilder


class TestSequentialBuilder(tf.test.TestCase):

    def test_input_names_and_shapes_full_model(self):
        defaults = {"type": "conv_block", "conv_dim": 3}
        builder = BlockBuilder(defaults=defaults)

        input_name = "test_input"
        input_shape = (4, 4, 4, 1)
        input_shapes = {input_name: input_shape}

        conv_layer_config = (
            {"filters": 3, },
            {"filters": 3, "pool_size": (2, 2, 2)},
            {"filters": 5, },
            {"type": "OutputCateg", "output_name": "ts_output", "categories": 2, "transition": "keras:Flatten"}
        )

        model = builder.build(input_shapes, conv_layer_config)
        self.assertEqual([input_name, ], model.input_names)
        self.assertEqual(input_shape, model.input_shape[1:])

    def test_attach_layer_conv(self):
        inp = layers.Input((6, 6, 1))
        defaults = {
            "type": "conv_block",
            "conv_dim": 2,
        }
        layer_config = {
            "filters": 2,
            "pool_size": 2,
            "dropout": 0.2,
            "batchnorm": True,
            "kernel_l2_reg": 0.0001
        }

        builder = BlockBuilder(defaults)
        x = builder.attach_block(inp, layer_config)
        model = Model(inp, x)

        self.assertIsInstance(model.layers[1], layers.Convolution2D)
        kreg = model.layers[1].get_config()["kernel_regularizer"]["config"]
        self.assertTrue("l1" not in kreg)
        self.assertAlmostEqual(kreg["l2"], layer_config["kernel_l2_reg"])

        self.assertIsInstance(model.layers[2], layers.BatchNormalization)
        self.assertIsInstance(model.layers[3], layers.Activation)
        self.assertIsInstance(model.layers[4], layers.MaxPooling2D)
        self.assertIsInstance(model.layers[5], layers.Dropout)
        self.assertEqual(model.output_shape[1:], (3, 3, 2))

    def test_attach_output_layers_regression_output_shape_and_names(self):
        config = {"type": "OutputRegErr", "output_names": ['out_A', 'out_B'], "flatten": True}

        inp = layers.Input((5, 1))
        builder = BlockBuilder()
        x = builder.attach_block(inp, config)
        model = Model(inp, x)

        target_output_shapes = {
            'out_A': (1,),
            'out_B': (1,),
            'out_A_err': (2,),
            'out_B_err': (2,),
        }

        output_shapes = {}
        for i, output_name in enumerate(model.output_names):
            output_shapes[output_name] = model.output_shape[i][1:]

        self.assertDictEqual(output_shapes, target_output_shapes)

    def test_attach_layer_dense(self):
        inp = layers.Input((3, 3, 1))
        defaults = {"type": "conv_block", "conv_dim": 2}
        layer_config = {"type": "dense_block", "units": 5, "dropout": 0.2, "batchnorm": True}

        builder = BlockBuilder(defaults)
        x = builder.attach_block(inp, layer_config)
        model = Model(inp, x)

        self.assertIsInstance(model.layers[1], layers.Dense)
        self.assertIsInstance(model.layers[2], layers.BatchNormalization)
        self.assertIsInstance(model.layers[3], layers.Activation)
        self.assertIsInstance(model.layers[4], layers.Dropout)
        self.assertEqual(model.output_shape[1:], (3, 3, 5))

    def test_attach_layer_wrong_layer_config_conv_block(self):
        inp = layers.Input((3, 3, 1))
        layer_config = {"type": "conv_block", "filters": 2, "units": 5}
        builder = BlockBuilder(None)

        with self.assertRaises(TypeError):
            builder.attach_block(inp, layer_config)

    def test_attach_layer_wrong_layer_config_dense_block(self):
        inp = layers.Input((3, 3, 1))
        layer_config = {"type": "dense_block", "filters": 2, "units": 5}
        builder = BlockBuilder(None)

        with self.assertRaises(TypeError):
            builder.attach_block(inp, layer_config)

    def test_attach_layer_cudnn_lstm(self):
        defaults = {"type": "conv_block", "conv_dim": 3, "recurrent_activation": "relu"}
        layer_config = {"type": "keras:LSTM", "units": 5, "activation": "relu"}
        builder = BlockBuilder(defaults)

        inp = layers.Input((10, 1))
        x = builder.attach_block(inp, layer_config)
        model = Model(inp, x)

        self.assertSequenceEqual(model.output_shape, (None, 5))
        check_these_in_config = {
            "units": 5,
            "recurrent_activation": "relu",
            "activation": "relu",
        }
        cfg = model.layers[-1].get_config()
        for k, v in check_these_in_config.items():
            self.assertEqual(cfg[k], v)



import tempfile
from unittest import TestCase
from unittest.mock import MagicMock, patch
import os
import shutil
import tensorflow as tf
from tensorflow.keras.models import Model
import tensorflow.keras.layers as layers

from orcanet.core import Organizer, Configuration, _extract_filepaths


class TestOrganizer(tf.test.TestCase):
    """
    Test the organizer by creating a dummy directory .temp/core.
    """
    def setUp(self):
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")
        os.mkdir(self.temp_dir)

        self.orga = Organizer(self.temp_dir)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    @patch('orcanet.core.tf.keras.utils.plot_model')
    def test_load_model_new_training(self, mock_plot_model):
        def plot_history(model, folder):
            pass
        mock_plot_model.side_effect = plot_history

        # latest epoch is None aka no model has been trained
        latest_epoch = None
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = error
        target_model = None
        with self.assertRaises(ValueError):
            self.orga._get_model(target_model, logging=False)

        # model given = ok
        target_model = build_test_model()
        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model.to_json(), model.to_json())

    def test_load_model_continue_training(self):
        # latest epoch is not None aka model has been trained before
        latest_epoch = (1, 1)
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = load model
        target_model = None
        model_file = os.path.join(self.temp_dir, "test_model.h5")
        self.orga.io.get_model_path = MagicMock(return_value=model_file)

        # no model given = ok
        saved_model = build_test_model()
        saved_model.save(model_file)
        loaded_model = self.orga._get_model(model=target_model, logging=False)
        self.assertEqual(saved_model.to_json(), loaded_model.to_json())

        # model given = ok
        target_model = 1.2

        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model, model)

    def test_val_is_due(self):
        no_of_files = 3
        self.orga.cfg.validate_interval = 2
        targets = {
            (1, 1): False,
            (1, 2): True,
            (1, 3): True,

            (2, 1): False,
            (2, 2): True,
        }

        self.orga.io.get_no_of_files = MagicMock(return_value=no_of_files)
        for epoch, target in targets.items():
            value = self.orga.val_is_due(epoch)
            self.assertEqual(value, target)

    def test_check_if_pred_already_done(self):
        epoch, fileno = 1, 1  # these dont matter
        # latest_prediction_file_no = None, total_files = 5 -> pred_done = False
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=None)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

        # latest_prediction_file_no = 2, total_files = 5 -> pred_done = False
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=2)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

        # latest_prediction_file_no = 4, total_files = 5 -> pred_done = True
        self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=2)
        self.orga.io.get_no_of_files = MagicMock(return_value=5)

        value = self.orga._check_if_pred_already_done(epoch, fileno)
        target = False
        self.assertEqual(value, target)

    # @patch('orcanet.core.make_model_prediction')  TODO with all the mocks this doesnt make sense anymore.
    # def test_predict(self, mocked):
    #     self.orga.cfg._list_file = "test/test.toml"
    #     func_args = []
    #
    #     def mock_make_model_prediction(orga, model, epoch, fileno, latest_val_file_no, samples=None):
    #         func_args.extend([orga, model, epoch, fileno, latest_val_file_no, samples])
    #
    #     mocked.side_effect = mock_make_model_prediction
    #
    #     self.orga.load_saved_model = MagicMock(return_value=build_test_model())
    #     self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=1)
    #     self.orga.io.get_no_of_files = MagicMock(return_value=1)
    #     self.orga.io.get_pred_files_list = MagicMock(return_value=
    #                                                    ['pred_model_epoch_2_file_2_on_bg_list_val_file_0.h5',
    #                                                     'pred_model_epoch_2_file_2_on_bg_list_val_file_1.h5'])
    #
    #     value_filepath = self.orga.predict(epoch=1, fileno=1)
    #     target_filepath = "./predictions/pred_model_epoch_1_file_1_on_test_val_files.h5"
    #
    #     self.assertEqual(value_filepath, target_filepath)
    #     self.assertEqual(func_args[2], value_filepath)
    #     self.assertEqual(func_args[3], None)


class TestZeroCenter(TestCase):
    def setUp(self):
        self.orga = Organizer("./")
        self.orga.cfg.zero_center_folder = "nope/"

    def test_get_xs_mean_existing(self):
        target = "asd"
        self.orga.xs_mean = target
        value = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, value)

    def test_get_xs_mean_no_folder(self):
        self.orga.cfg.zero_center_folder = None
        with self.assertRaises(ValueError):
            self.orga.get_xs_mean(logging=False)

    @patch('orcanet.core.load_zero_center_data')
    def test_get_xs_mean_new(self, load_mock):
        target = "blub"

        def mock_xs_mean():
            return target

        load_mock.return_value = mock_xs_mean()
        xs_mean = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, xs_mean)
        self.assertEqual(target, self.orga.xs_mean)


class TestConfiguration(TestCase):
    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.working_list = self.data_folder + "/in_out_test_list.toml"
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")
        os.mkdir(self.temp_dir)

    def tearDown(self):
        os.rmdir(self.temp_dir)

    def test_output_folder_bracket(self):
        list_file = None
        config_file = None

        output_folder = "test"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder + "/", cfg.output_folder)

        output_folder = "test/"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder, cfg.output_folder)

    def test_listfile_already_given(self):
        config_file = None
        output_folder = "test"

        list_file = self.working_list
        cfg = Configuration(output_folder, list_file, config_file)

        with self.assertRaises(ValueError):
            cfg.import_list_file(list_file)

    def test_get_files_not_given(self):
        cfg = Configuration("./", None, None)
        with self.assertRaises(AttributeError):
            cfg.get_files("train")
        with self.assertRaises(AttributeError):
            cfg.get_files("val")
        with self.assertRaises(AttributeError):
            cfg.get_files("inference")
        with self.assertRaises(NameError):
            cfg.get_files("schmu")

    def test_update_config_bad_format(self):
        cfg = Configuration("./", None, None)

        content = [
            "[config]",
            "schmu=1",
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(AttributeError):
                cfg.update_config(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_ok(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',
            'inf = ["input_A_inf_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation = ["input_B_val_1.h5",]',
            'inference_files = ["input_B_inf_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            cfg.import_list_file(file_name)

        finally:
            os.remove(file_name)

    def test_load_list_file_unknown_keyword(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',
            'schmu = 1',
            
            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_no_train_files(self):
        cfg = Configuration("./", None, None)
        val_files = {"input_A": ("input_A_val_1.h5", ),
                     "input_B": ("input_B_val_1.h5", )}
        content = [
            '[input_A]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            cfg.import_list_file(file_name)
            with self.assertRaises(AttributeError):
                cfg.get_files("train")
            with self.assertRaises(AttributeError):
                cfg.get_files("inference")

            self.assertEqual(val_files, cfg.get_files("val"))

        finally:
            os.remove(file_name)

    def test_load_list_file_no_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'schmu=1',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_train_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", ]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation = ["input_A_val_1.h5", "input_A_val_2.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")

            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_extract_filepaths_dir_given_in_toml(self):
        with tempfile.TemporaryDirectory() as tempdir:
            h5files = tuple([
                os.path.join(tempdir, fname)
                for fname in ("file1.h5", "file2.h5")
            ])
            for fname in h5files:
                open(os.path.join(tempdir, fname), "w").close()
            content = {
                'input_A':
                    {'train_files': [tempdir]},
            }
            result = _extract_filepaths(content, "train_files")["input_A"]
            self.assertTupleEqual(result, h5files)


def build_test_model(compile=False):
    input_shape = (1,)
    inp = layers.Input(input_shape, name="inp")
    outp = layers.Dense(1, name="out")(inp)

    test_model = Model(inp, outp)
    if compile:
        test_model.compile("sgd", loss={"out": "mse"}, metrics={"out": "mae"})

    return test_model



# -*- coding: utf-8 -*-
"""
Visualization tools used without Keras.
Makes performance graphs for training and validating.
"""
import os
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages


class TrainValPlotter:
    """
    Class for plotting train/val curves.

    Instructions
    ------------
    1. Use tvp.plot_curves(train, val) once or more on pairs of
        train/val data.
    2. When all lines are plotted, use tvp.apply_layout() once for proper
        scaling, ylims, etc.

    """
    def __init__(self):
        # White space added below and above points
        self.y_lim_padding = [0.10, 0.25]
        # Store all plotted points for setting x/y lims
        self._xpoints_train = np.array([])
        self._xpoints_val = np.array([])
        self._ypoints_train = np.array([])
        self._ypoints_val = np.array([])

    def plot_curves(self,
                    train_data,
                    val_data=None,
                    train_label="training",
                    val_label="validation",
                    color=None,
                    smooth_sigma=None,
                    tlw=0.5,
                    vlw=0.5,
                    vms=3):
        """
        Plot a training and optionally a validation line.

        The data can contain nan's.

        Parameters
        ----------
        train_data : List
            X data [0] and y data [1] of the train curve. Will be plotted as
            connected dots.
        val_data : List, optional
            Optional X data [0] and y data [1] of the validation curve.
            Will be plotted as a faint solid line of the same color as train.
        train_label : str, optional
            Label for the train line in the legend.
        val_label : str, optional
            Label for the validation line in the legend.
        color : str, optional
            Color used for the train/val line.
        smooth_sigma : int, optional
            Apply gaussian blur to the train curve with given sigma.
        tlw : float
            Linewidth of train curve.
        vlw : float
            Linewidth of val curve.
        vms : float
            Markersize of the val curve.

        """
        if train_data is None and val_data is None:
            raise ValueError(
                "Can not plot when no train and val data is given.")

        if train_data is not None:
            epoch, y_data = train_data
            if smooth_sigma is not None:
                y_data = gaussian_smooth(y_data, smooth_sigma)

            self._xpoints_train = np.concatenate((self._xpoints_train, epoch))
            self._ypoints_train = np.concatenate((self._ypoints_train, y_data))

            train_plot = plt.plot(
                epoch, y_data, color=color, ls='-',
                zorder=3, label=train_label, lw=tlw, alpha=0.5)
            train_color = train_plot[0].get_color()
        else:
            train_color = color

        if val_data is not None:
            self._xpoints_val = np.concatenate((self._xpoints_val,
                                                val_data[0]))
            self._ypoints_val = np.concatenate((self._ypoints_val,
                                                val_data[1]))

            val_data_clean = skip_nans(val_data)
            # val plot always has the same color as the train plot
            plt.plot(val_data_clean[0], val_data_clean[1], color=train_color,
                     marker='o', zorder=3, lw=vlw, markersize=vms, label=val_label)

    def apply_layout(self,
                     title=None,
                     x_label="Epoch",
                     y_label=None,
                     grid=True,
                     legend=True,
                     x_lims=None,
                     y_lims="auto",
                     x_ticks="auto",
                     logy=False):
        """
        Apply given layout.
        Can calculate good y_lims and x_ticks automatically.

        Parameters
        ----------
        title : str
            Title of the plot.
        x_label : str
            X label of the plot.
        y_label : str
            Y label of the plot.
        grid : bool
            If true, show a grid.
        legend : bool
            If true, show a legend.
        x_lims : List
            X limits of the data.
        y_lims : List or str
            Y limits of the data. "auto" for auto-calculation.
        x_ticks : List
            Positions of the major x ticks.
        logy : bool
            If true, make y axis log.

        """
        if logy:
            plt.yscale("log")
        if x_ticks is not None:
            if x_ticks == "auto":
                all_x_points = np.concatenate(
                    (self._xpoints_train, self._xpoints_val)
                )
                x_ticks = get_epoch_xticks(all_x_points)
            else:
                x_ticks = x_ticks
            plt.xticks(x_ticks)

        if x_lims is not None:
            plt.xlim(x_lims)

        if y_lims is not None:
            if y_lims == "auto":
                y_lims = get_ylims(
                    self._ypoints_train,
                    self._ypoints_val,
                    fraction=self.y_lim_padding,
                )
            else:
                y_lims = y_lims
            plt.ylim(y_lims)

        if legend:
            plt.legend(loc='upper right')

        plt.xlabel(x_label)
        plt.ylabel(y_label)

        if title is not None:
            title = plt.title(title)
            title.set_position([.5, 1.04])

        if grid:
            plt.grid(True, zorder=0, linestyle='dotted')


def gaussian_smooth(y, sigma, truncate=4):
    """ Smooth a 1d ndarray with a gaussian filter. """
    # kernel_width = 2 * sigma * truncate + 1
    kernel_x = np.arange(-truncate * sigma, truncate * sigma + 1)
    kernel = _gauss(kernel_x, 0, sigma)
    y = np.pad(np.asarray(y), int(len(kernel)/2), "edge")
    blurred = np.convolve(y, kernel, "valid")
    return blurred


def _gauss(x, mu=0, sigma=1):
    return (1/(np.sqrt(2*np.pi)*sigma)) * np.exp(-np.power(x - mu, 2.) / (2 * np.power(sigma, 2.)))


def plot_history(train_data,
                 val_data=None,
                 train_label="training",
                 val_label="validation",
                 color=None,
                 **kwargs):
    """
    Plot the train/val curves in a single plot.

    For backward compat. Functionality moved to TrainValPlotter

    """
    tvp = TrainValPlotter()
    tvp.plot_curves(train_data,
                    val_data,
                    train_label=train_label,
                    val_label=val_label,
                    color=color)
    tvp.apply_layout(**kwargs)


def skip_nans(data):
    """
    Skip over nan values, so that all dots are connected.

    Parameters
    ----------
    data : List
        Contains x and y data as ndarrays. The y values may contain nans.

    Returns
    -------
    data_clean : List
        Contains x and y data as ndarrays. Points with y=nan are skipped.

    """
    not_nan = ~np.isnan(data[1])
    data_clean = data[0][not_nan], data[1][not_nan]
    return data_clean


def get_ylims(y_points_train, y_points_val=None, fraction=0.25):
    """
    Get the y limits for the summary plot.

    For the training data, limits are calculated while ignoring data points
    which are far from the median (in terms of the median distance
    from the median).
    This is because there are outliers sometimes in the training data,
    especially early on in the training.

    Parameters
    ----------
    y_points_train : List
        y data of the train curve.
    y_points_val : List or None
        Y data of the validation curve.
    fraction : float or List
        How much whitespace of the total y range is added below and above
        the lines.

    Returns
    -------
    y_lims : tuple
        Minimum, maximum of the data.

    """
    assert not (y_points_train is None and
                y_points_val is None), "train and val data are None"

    def reject_outliers(data, threshold):
        d = np.abs(data - np.median(data))
        mdev = np.median(d)
        s = d / mdev if mdev else 0.
        no_outliers = data[s < threshold]
        lims = np.amin(no_outliers), np.amax(no_outliers)
        return lims

    mins, maxs = [], []
    if y_points_train is not None and len(y_points_train) != 0:
        y_train = y_points_train[~np.isnan(y_points_train)]
        y_lims_train = reject_outliers(y_train, 5)
        mins.append(y_lims_train[0])
        maxs.append(y_lims_train[1])

    if y_points_val is not None and len(y_points_val) != 0:
        y_val = y_points_val[~np.isnan(y_points_val)]

        if len(y_val) == 1:
            y_lim_val = y_val[0], y_val[0]
        else:
            y_lim_val = np.amin(y_val), np.amax(y_val)

        mins.append(y_lim_val[0])
        maxs.append(y_lim_val[1])

    if len(mins) == 1:
        y_lims = (mins[0], maxs[0])
    else:
        y_lims = np.amin(mins), np.amax(maxs)

    if y_lims[0] == y_lims[1]:
        y_range = 0.1 * y_lims[0]
    else:
        y_range = y_lims[1] - y_lims[0]

    try:
        fraction = float(fraction)
        padding = [fraction, fraction]
    except TypeError:
        # is a list
        padding = fraction

    if padding != [0., 0.]:
        y_lims = (y_lims[0] - padding[0] * y_range,  y_lims[1] + padding[1] * y_range)

    return y_lims


def get_epoch_xticks(x_points):
    """
    Calculates the xticks for the train and validation summary plot.

    One tick per epoch. Less the larger #epochs is.

    Parameters
    ----------
    x_points : List
        A list of the x coordinates of all points.

    Returns
    -------
    x_ticks_major : numpy.ndarray
        Array containing the ticks.

    """
    if len(x_points) == 0:
        raise ValueError("x-coordinates are empty!")

    minimum, maximum = np.amin(x_points), np.amax(x_points)
    if maximum - minimum > 0.5:
        # for longer trainings
        start_epoch, end_epoch = np.floor(minimum), np.ceil(maximum)
        # less xticks if there are many epochs
        n_epochs = end_epoch - start_epoch
        x_ticks_stepsize = 1 + np.floor(n_epochs / 20.)
        x_ticks_major = np.arange(
            start_epoch, end_epoch + x_ticks_stepsize, x_ticks_stepsize)
    else:
        # for early peeks
        start_epoch = np.floor(minimum)
        end_epoch = maximum + minimum - start_epoch
        x_ticks_major = np.linspace(
            start_epoch, end_epoch, 6)

    return x_ticks_major


def update_summary_plot(orga):
    """
    Plot and save all metrics of the given validation- and train-data
    into a pdf file, each metric in its own plot.

    If metric pairs of a variable and its error are found (e.g. e_loss
    and e_err_loss), they will have the same color and appear back to
    back in the plot.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.

    """
    plt.ioff()
    pdf_name = orga.io.get_subfolder("plots", create=True) + "/summary_plot.pdf"

    # Extract the names of the metrics
    all_metrics = orga.history.get_metrics()
    # Sort them
    all_metrics = sort_metrics(all_metrics)
    # Plot them w/ custom color cycle
    colors = ['#000000', '#332288', '#88CCEE', '#44AA99', '#117733', '#999933',
              '#DDCC77', '#CC6677', '#882255', '#AA4499', '#661100', '#6699CC',
              '#AA4466', '#4477AA']  # ref. personal.sron.nl/~pault/
    color_counter = 0
    with PdfPages(pdf_name) as pdf:
        for metric_no, metric in enumerate(all_metrics):
            # If this metric is an err metric of a variable, color it the same
            if all_metrics[metric_no-1] == metric.replace("_err", ""):
                color_counter -= 1
            orga.history.plot_metric(
                metric, color=colors[color_counter % len(colors)])
            plt.suptitle(
                os.path.basename(os.path.abspath(orga.cfg.output_folder)))
            color_counter += 1
            pdf.savefig()
            plt.clf()

        orga.history.plot_lr()
        color_counter += 1
        pdf.savefig()
        plt.close()


def sort_metrics(metric_names):
    """
    Sort a list of metrics, so that errors are right after their variable.
    The format of the metric names have to be e.g. e_loss and e_err_loss
    for this to work.

    Example
    ----------
    >>> sort_metrics( ['e_loss', 'loss', 'e_err_loss', 'dx_err_loss'] )
    ['e_loss', 'e_err_loss', 'loss', 'dx_err_loss']

    Parameters
    ----------
    metric_names : List
        List of metric names.

    Returns
    -------
    sorted_metrics : List
        List of sorted metric names with the same length as the input.

    """
    sorted_metrics = [0] * len(metric_names)
    counter = 0
    for metric_name in metric_names:
        if "err_" in metric_name:
            if metric_name.replace("err_", "") not in metric_names:
                sorted_metrics[counter] = metric_name
                counter += 1
            continue
        sorted_metrics[counter] = metric_name
        counter += 1
        err_loss = metric_name.split("_loss")[0]+"_err_loss"
        if err_loss in metric_names:
            sorted_metrics[counter] = err_loss
            counter += 1

    assert 0 not in sorted_metrics, "Something went wrong with the sorting of " \
                                    "metrics! Given was {}, output was " \
                                    "{}. ".format(metric_names, sorted_metrics)

    return sorted_metrics



from unittest import TestCase
import numpy as np
import orcanet.lib.dataset_modifiers as dmods


class TestFunctions(TestCase):
    def test_as_array(self):
        y_values = "y_values gets simply passed forward"
        y_true = {
            "out_A": 1,
            "out_B": 2,
        }
        y_pred = {
            "out_pred_A": 3,
            "out_pred_B": 4,
        }
        info_blob = {
            "y_values": y_values,
            "ys": y_true,
            "y_pred": y_pred,
        }

        target = {
            "y_values": y_values,
            "label_out_A": 1,
            "label_out_B": 2,
            "pred_out_pred_A": 3,
            "pred_out_pred_B": 4,
        }

        datasets = dmods.as_array(info_blob)
        self.assertDictEqual(datasets, target)


class TestAsRecarrayDist(TestCase):
    def setUp(self):
        inp = {
            "y_pred": {"aa": np.ones((5, 2)), "bb": np.ones((5, 2, 3))*2},
            "ys": {"aa": np.ones((5, 2))*3, "bb": np.ones((5, 2, 3))*4},
        }
        self.output = dmods.as_recarray_dist(inp)

    def test_pred_shape(self):
        self.assertTrue(self.output["pred"].shape == (5, ))

    def test_pred_names(self):
        self.assertTupleEqual(
            self.output["pred"].dtype.names,
            ('aa_1', 'aa_err_1', 'bb_1', 'bb_2', 'bb_3', 'bb_err_1', 'bb_err_2', 'bb_err_3'))

    def test_pred_array_1(self):
        np.testing.assert_array_equal(self.output["pred"]["aa_1"], np.ones(5))

    def test_pred_array_2(self):
        np.testing.assert_array_equal(self.output["pred"]["bb_1"], np.ones(5)*2)

    def test_true_shape(self):
        self.assertTrue(self.output["true"].shape == (5, ))

    def test_true_names(self):
        self.assertTupleEqual(
            self.output["true"].dtype.names, ('aa_1', 'bb_1', 'bb_2', 'bb_3'))

    def test_true_array_1(self):
        np.testing.assert_array_equal(self.output["true"]["aa_1"], np.ones(5)*3)

    def test_true_array_2(self):
        np.testing.assert_array_equal(self.output["true"]["bb_1"], np.ones(5)*4)


class TestAsRecarrayDistSplit(TestAsRecarrayDist):
    def setUp(self):
        inp = {
            "y_pred": {
                "aa": np.ones(5,)*5, "aa_err": np.ones((5, 2)),
                "bb": np.ones((5, 3))*6, "bb_err": np.ones((5, 2, 3))*2,
            },
            "ys": {
                "aa": np.ones(5,)*7, "aa_err": np.ones((5, 2))*3,
                "bb": np.ones((5, 3))*8, "bb_err": np.ones((5, 2, 3))*4,
            },
        }
        self.output = dmods.as_recarray_dist_split(inp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility code regarding user input.
"""

import os
import shutil
import h5py
import numpy as np
from inspect import signature

# moved into IOHandler.get_batch for speed up; tensorflow import is slow!
# from orcanet.h5_generator import Hdf5BatchGenerator


def get_subfolder(main_folder, name=None, create=False):
    """
    Get the path to one or all subfolders of the main folder.

    Parameters
    ----------
    main_folder : str
        The main folder.
    name : str or None
        The name of the subfolder.
    create : bool
        If the subfolder should be created if it does not exist.

    Returns
    -------
    subfolder : str or tuple
        The path of the subfolder. If name is None, all subfolders
        will be returned as a tuple.

    """
    if not main_folder[-1] == "/":
        main_folder += "/"

    subfolders = {
        "train_log": main_folder + "train_log",
        "saved_models": main_folder + "saved_models",
        "plots": main_folder + "plots",
        "activations": main_folder + "plots/activations",
        "predictions": main_folder + "predictions",
        "inference": main_folder + "predictions/inference"
    }

    def get(fdr):
        subfdr = subfolders[fdr]
        if create and not os.path.exists(subfdr):
            print("Creating directory: " + subfdr)
            os.makedirs(subfdr)
        return subfdr

    if name is None:
        subfolder = [get(name) for name in subfolders]
    else:
        subfolder = get(name)
    return subfolder


def get_inputs(model):
    """ Get names and keras layers of the inputs of the model, as a dict.
    """
    return {name: model.get_layer(name) for name in model.input_names}


class IOHandler(object):
    """
    Access info indirectly contained in the cfg object.
    """
    def __init__(self, cfg):
        self.cfg = cfg

        # copies of files on local tmpdir
        self._tmpdir_files_dict = {
            "train": None,
            "val": None,
            "inference": None,
        }

    def get_latest_epoch(self):
        """
        Return the highest epoch/fileno pair of any saved model.

        Returns
        -------
        latest_epoch : tuple or None
            The highest epoch, file_no pair. None if the folder is
            empty or does not exist yet.

        """
        epochs = self.get_all_epochs()
        if len(epochs) == 0:
            latest_epoch = None
        else:
            latest_epoch = epochs[-1]

        return latest_epoch

    def get_all_epochs(self):
        """
        Get a sorted list of the epoch/fileno pairs of all saved models.

        Returns
        -------
        epochs : List
            The (epoch, fileno) tuples. List is empty if none can be found.
        """
        saved_models_folder = self.cfg.output_folder + "saved_models"
        epochs = []

        if os.path.exists(saved_models_folder):
            files = []
            for file in os.listdir(saved_models_folder):
                if file.startswith("model_epoch_") and file.endswith('.h5'):
                    files.append(file)

            for file in files:
                # model_epoch_XX_file_YY
                file_base = os.path.splitext(file)[0]
                f_epoch, file_no = file_base.split(
                    "model_epoch_")[-1].split("_file_")
                epochs.append((int(f_epoch), int(file_no)))
            epochs.sort()

        return epochs

    def get_next_epoch(self, epoch):
        """
        Return the next epoch / fileno tuple.

        It depends on how many train files there are.

        Parameters
        ----------
        epoch : tuple or None
            Current epoch and file number.

        Returns
        -------
        next_epoch : tuple
            Next epoch and file number.

        """
        if epoch is None:
            next_epoch = (1, 1)
        elif epoch[1] == self.get_no_of_files("train"):
            next_epoch = (epoch[0] + 1, 1)
        else:
            next_epoch = (epoch[0], epoch[1] + 1)
        return next_epoch

    def get_previous_epoch(self, epoch):
        """ Return the previous epoch / fileno tuple. """
        if epoch[1] == 1:
            if epoch[0] == 1:
                raise ValueError("Can not get previous epoch of epoch {} file {}".format(*epoch))
            n_train_files = self.get_no_of_files("train")
            prev_epoch = (epoch[0]-1, n_train_files)

        else:
            prev_epoch = (epoch[0], epoch[1]-1)

        return prev_epoch

    def get_subfolder(self, name=None, create=False):
        """
        Get the path to one or all subfolders of the main folder.

        Parameters
        ----------
        name : str or None
            The name of the subfolder.
        create : bool
            If the subfolder should be created if it does not exist.

        Returns
        -------
        subfolder : str or tuple
            The path of the subfolder. If name is None, all subfolders
            will be returned as a tuple.

        """
        subfolder = get_subfolder(self.cfg.output_folder, name, create)
        return subfolder

    def get_model_path(self, epoch, fileno, local=False):
        """
        Get the path to a model (which might not exist yet).

        Parameters
        ----------
        epoch : int
            Its epoch.
        fileno : int
            Its file number.
        local : bool
            If True, will only return the path inside the output_folder,
            i.e. models/models_epochXX_file_YY.h5.

        Returns
        -------
        model_path : str
            The path to the model.
        """
        if epoch == -1 and fileno == -1:
            epoch, fileno = self.get_latest_epoch()
        if epoch < 1 or fileno < 1:
            raise ValueError("Invalid epoch/file number {}, {}: Must be "
                             "either (-1, -1) or both >0".format(epoch, fileno))

        subfolder = self.get_subfolder("saved_models")
        if local:
            subfolder = subfolder.split("/")[-1]
        file_name = 'model_epoch_{}_file_{}.h5'.format(epoch, fileno)

        model_path = subfolder + "/" + file_name
        return model_path

    def get_latest_prediction_file_no(self, epoch, fileno):
        """
        Returns the file number of the latest currently predicted val file.

        Parameters
        ----------
        epoch : int
            Epoch of the model that has predicted.
        fileno : int
            Fileno of the model that has predicted.

        Returns
        -------
        latest_val_file_no : int or None
            File number of the prediction file with the highest val index.
            STARTS FROM 1, so this is whats in the file name.
            None if there is none.

        """
        prediction_folder = self.get_subfolder("predictions", create=True)

        val_file_nos = []

        for file in os.listdir(prediction_folder):
            # name e.g.: pred_model_epoch_6_file_1_on_list_val_file_1.h5
            if not (file.endswith(".h5") and file.startswith("pred_model")):
                continue

            f_epoch, f_fileno, val_file_no = split_name_of_predfile(file)
            if f_epoch == epoch and f_fileno == fileno:
                val_file_nos.append(val_file_no)

        if len(val_file_nos) == 0:
            latest_val_file_no = None
        else:
            latest_val_file_no = max(val_file_nos)

        return latest_val_file_no

    def get_pred_path(self, epoch, fileno, pred_file_no):
        """
        Gets the path of a prediction file. The ints all start from 1.

        Parameters
        ----------
        epoch : int
            Epoch of an already trained nn model.
        fileno : int
            File number train step of an already trained nn model.
        pred_file_no : int
            Val file no of the prediction files that are found in the
            prediction folder.

        Returns
        -------
        pred_filepath : str
            The path.

        """
        list_file = self.cfg.get_list_file()
        if list_file is None:
            raise ValueError("No toml list file specified. Can not look up "
                             "saved prediction")
        list_name = os.path.splitext(os.path.basename(list_file))[0]

        pred_filepath = self.get_subfolder("predictions") + \
            '/pred_model_epoch_{}_file_{}_on_{}_val_file_{}.h5'.format(
                epoch, fileno, list_name, pred_file_no)

        return pred_filepath

    def get_pred_files_list(self, epoch=None, fileno=None):
        """
        Returns a sorted list with all pred .h5 files in the prediction folder.
        Does not include the inference files.

        Parameters
        ----------
        epoch : int, optional
            Specific model epoch to look pred files up for.
        fileno : int, optional
            Specific model epoch to look pred files up for.

        Returns
        -------
        pred_files_list : List
            List with the full filepaths of all prediction results files.

        """
        prediction_folder = self.get_subfolder("predictions")

        pred_files_list = []
        for file in os.listdir(prediction_folder):
            if not (file.startswith("pred_model_epoch") and file.endswith(".h5")):
                continue
            pred_file = os.path.join(prediction_folder, file)
            p_epoch, p_file_no, p_val_file_no = split_name_of_predfile(pred_file)
            if epoch is not None and epoch != p_epoch:
                continue
            if fileno is not None and fileno != p_file_no:
                continue
            pred_files_list.append(pred_file)

        pred_files_list.sort()  # sort predicted val files from 1 ... n
        return pred_files_list

    def get_local_files(self, which):
        """
        Get the training or validation file paths for each list input set.

        Returns the path to the copy of the file on the local tmpdir, which
        it will generate if called for the first time.

        Parameters
        ----------
        which : str
            Either "train", "val", or "inference".

        Returns
        -------
        dict
            A dict containing the paths to the training or validation files on
            which the model will be trained on. Example for the format for
            two input sets with two files each:
            {
             "input_A" : ('path/to/set_A_file_1.h5', 'path/to/set_A_file_2.h5'),
             "input_B" : ('path/to/set_B_file_1.h5', 'path/to/set_B_file_2.h5'),
            }

        """
        if which not in self._tmpdir_files_dict.keys():
            raise NameError("Unknown fileset name ", which)

        files = self.cfg.get_files(which)
        if self.cfg.use_scratch_ssd:
            if self._tmpdir_files_dict[which] is None:
                self._tmpdir_files_dict[which] = use_local_tmpdir(files)
            return self._tmpdir_files_dict[which]
        else:
            return files

    def get_n_bins(self):
        """
        Get the number of bins from the training files.

        Only the first files are looked up, the others should be identical.

        Returns
        -------
        n_bins : dict
            Toml-list input names as keys, list of the bins as values.

        """
        # TODO check if bins are equal in all files?
        train_files = self.get_local_files("train")
        n_bins = {}
        for input_key in train_files:
            with h5py.File(train_files[input_key][0], "r") as f:
                n_bins[input_key] = f[self.cfg.key_x_values].shape[1:]
        return n_bins

    def get_file_sizes(self, which):
        """
        Get the number of samples in each training or validation input file.

        Parameters
        ----------
        which : str
            Either train or val.

        Returns
        -------
        file_sizes : List
            Its length is equal to the number of files in each input set.

        Raises
        ------
        ValueError
            If there is a different number of samples in any of the
            files of all inputs.

        """
        file_sizes_full, error_file_sizes, file_sizes = {}, [], []
        for n, file_no_set in enumerate(self.yield_files(which)):
            # the number of samples in the n-th file of all inputs
            file_sizes_full[n] = [h5_get_number_of_rows(
                file, datasets=[self.cfg.key_y_values, self.cfg.key_x_values])
                for file in file_no_set.values()]
            if not file_sizes_full[n].count(file_sizes_full[n][0]) == \
                    len(file_sizes_full[n]):
                    error_file_sizes.append(n)
            else:
                file_sizes.append(file_sizes_full[n][0])

        if len(error_file_sizes) != 0:
            err_msg = "The files you gave for the different inputs of the model " \
                      "do not all have the same number of samples!\n"
            for n in error_file_sizes:
                err_msg += "File no {} in {} has the following files sizes " \
                           "for the different inputs: {}\n".format(
                            n, which, file_sizes_full[n])
            raise ValueError(err_msg)

        return file_sizes

    def get_no_of_files(self, which):
        """
        Return the number of training or validation files.

        Only looks up the no of files of one (random) list input, as equal
        length is checked during read in.

        Parameters
        ----------
        which : str
            Either train or val.

        Returns
        -------
        no_of_files : int
            The number of files.

        """
        files = self.get_local_files(which)
        no_of_files = len(list(files.values())[0])
        return no_of_files

    def yield_files(self, which):
        """
        Yield a training or validation filepaths for every input.

        They will be yielded in the same order as they are given in the
        toml file.

        Parameters
        ----------
        which : str
            Either train or val.

        Yields
        ------
        files_dict : dict
            Keys: The name of every toml list input.
            Values: One of the filepaths.

        """
        files = self.get_local_files(which)
        for file_no in range(self.get_no_of_files(which)):
            files_dict = {key: files[key][file_no] for key in files}
            yield files_dict

    def get_file(self, which, file_no):
        """ Get a dict with the n-th files. """
        files = self.get_local_files(which)
        files_dict = {key: files[key][file_no-1] for key in files}
        return files_dict

    def check_connections(self, model):
        """
        Check if the names and shapes of the samples and labels in the
        given input files work with the model.

        Also takes into account the possibly present sample or label modifiers.

        Parameters
        ----------
        model : ks.model
            A keras model.

        Raises
        ------
        ValueError
            If they dont work together.

        """
        print("\nInput check\n-----------")
        # Get a batch of data to investigate the given modifier functions
        info_blob = self.get_batch()
        y_values = info_blob["y_values"]
        layer_inputs = get_inputs(model)
        # keys: name of layers, values: shape of input
        layer_inp_shapes = {key: layer_inputs[key].input_shape[0][1:]
                            for key in layer_inputs}
        list_inp_shapes = self.get_n_bins()

        print("The data in the files of the toml list have the following "
              "names and shapes:")
        for list_key in list_inp_shapes:
            print("\t{}\t{}".format(list_key, list_inp_shapes[list_key]))

        if self.cfg.sample_modifier is None:
            print("\nYou did not specify a sample modifier.")
            info_blob["xs"] = info_blob["x_values"]
        else:
            modified_xs = self.cfg.sample_modifier(info_blob)
            modified_shapes = {
                modi_key: tuple(modified_xs[modi_key].shape)[1:]
                for modi_key in modified_xs
            }
            print("\nAfter applying your sample modifier, they have the "
                  "following names and shapes:")
            for list_key in modified_shapes:
                print("\t{}\t{}".format(list_key, modified_shapes[list_key]))
            list_inp_shapes = modified_shapes
            info_blob["xs"] = modified_xs

        print("\nYour model requires the following input names and shapes:")
        for layer_key in layer_inp_shapes:
            print("\t{}\t{}".format(layer_key, layer_inp_shapes[layer_key]))

        # Both inputs are dicts with  name: shape  of input/output layers/data
        err_inp_names, err_inp_shapes = [], []
        for layer_name in layer_inp_shapes:
            if layer_name not in list_inp_shapes.keys():
                # no matching name
                err_inp_names.append(layer_name)
            elif list_inp_shapes[layer_name] != layer_inp_shapes[layer_name]:
                # no matching shape
                err_inp_shapes.append(layer_name)

        err_msg_inp = ""
        if len(err_inp_names) == 0 and len(err_inp_shapes) == 0:
            print("\nInput check passed.")
        else:
            print("\nInput check failed!")
            if len(err_inp_names) != 0:
                err_msg_inp += "No matching input name from the input files " \
                               "for input layer(s): " + (
                                ", ".join(str(e) for e in err_inp_names) + "\n")
            if len(err_inp_shapes) != 0:
                err_msg_inp += "Shapes of layers and labels do not match for " \
                               "the following input layer(s): " + (
                                ", ".join(str(e) for e in err_inp_shapes) + "\n")
            print("Error:", err_msg_inp)

        # ----------------------------------
        print("\nOutput check\n------------")
        # tuple of strings
        mc_names = y_values.dtype.names
        print("The following {} label names are in the first file of the "
              "toml list:".format(len(mc_names)))
        print("\t" + ", ".join(str(name) for name in mc_names), end="\n\n")

        if self.cfg.label_modifier is not None:
            label_names = tuple(self.cfg.label_modifier(info_blob).keys())
            print("The following {} labels get produced from them by your "
                  "label_modifier:".format(len(label_names)))
            print("\t" + ", ".join(str(name) for name in label_names), end="\n\n")
        else:
            label_names = mc_names
            print("You did not specify a label_modifier. The output layers "
                  "will be provided with labels that match their name from "
                  "the above.\n\n")

        # tuple of strings
        loss_names = tuple(model.output_names)
        print("Your model has the following {} output layers:".format(
            len(loss_names)))
        print("\t" + ", ".join(str(name) for name in loss_names), end="\n\n")

        err_out_names = []
        for loss_name in loss_names:
            if loss_name not in label_names:
                err_out_names.append(loss_name)

        err_msg_out = ""
        if len(err_out_names) == 0:
            print("Output check passed.\n")
        else:
            print("Output check failed!")
            if len(err_out_names) != 0:
                err_msg_out += "No matching label name from the input files " \
                               "for output layer(s): " + (
                                ", ".join(str(e) for e in err_out_names) + "\n")
            print("Error:", err_msg_out)

        err_msg = err_msg_inp + err_msg_out
        if err_msg != "":
            raise ValueError(err_msg)

    def get_batch(self):
        """
        For testing purposes, return a batch of x_values and y_values.

        This will always be the first batchsize samples and y_values from
        the first file, before any modifiers have been applied.

        Returns
        -------
        info_blob : dict
            X- and y-values from the files. Has the following entries:
            x_values : dict
                Keys: Names of the input datasets from the list toml file.
                Values: ndarray, a batch of samples.
            y_values : ndarray
                From the y_values datagroup of the input files.

        """
        # this will import tf; move inside here for speed up
        from orcanet.h5_generator import Hdf5BatchGenerator

        gen = Hdf5BatchGenerator(
            next(self.yield_files("train")),
            batchsize=self.cfg.batchsize,
            key_x_values=self.cfg.key_x_values,
            key_y_values=self.cfg.key_y_values,
            keras_mode=False,
        )
        info_blob = gen[0]
        info_blob.pop("xs")
        info_blob.pop("ys")
        return info_blob

    def get_input_shapes(self):
        """
        Get the input names and shapes of the data after the modifier has
        been applied.

        Returns
        -------
        input_shapes : dict
            Keys: Name of the inputs of the model.
            Values: Their shape without the batchsize.

        """
        if self.cfg.sample_modifier is None:
            input_shapes = self.get_n_bins()
        else:
            info_blob = self.get_batch()
            xs_mod = self.cfg.sample_modifier(info_blob)
            input_shapes = {
                input_name: tuple(input_xs.shape)[1:]
                for input_name, input_xs in xs_mod.items()
            }
        return input_shapes

    def print_log(self, lines, logging=True):
        """ Print and also log to the full log file. """
        if isinstance(lines, str):
            lines = [lines, ]

        if not logging:
            for line in lines:
                print(line)
        else:
            full_log_file = self.cfg.output_folder + 'log.txt'
            with open(full_log_file, 'a+') as f_out:
                for line in lines:
                    f_out.write(line + "\n")
                    print(line)

    def get_epoch_float(self, epoch, fileno):
        """ Make a float value out of epoch/fileno. """
        # calculate the fraction of samples per file compared to all files,
        # e.g. [100, 50, 50] --> [0.5, 0.75, 1]
        file_sizes = self.get_file_sizes("train")
        file_sizes_rltv = np.cumsum(file_sizes) / np.sum(file_sizes)

        epoch_float = epoch - 1 + file_sizes_rltv[fileno - 1]
        return epoch_float

    def get_learning_rate(self, epoch):
        """
        Get the learning rate for a given epoch and file number.

        The user learning rate (cfg.learning_rate) can be None, a float,
        a tuple, or a function.

        Parameters
        ----------
        epoch : tuple
            Epoch and file number. Both start at 1, i.e. the start of the
            training is (1, 1), the next file is (1, 2), ...
            This is also in the filename of the saved models.

        Returns
        -------
        lr : float
            The learning rate that will be used for the given epoch/fileno.

        """
        error_msg = "The learning rate must be either a float, a tuple of " \
                    "two floats or a function."
        no_train_files = self.get_no_of_files("train")
        user_lr = self.cfg.learning_rate

        if isinstance(user_lr, str):
            # read lr from a csv file in the main folder, which must have
            # 3 columns (Epoch, fileno, lr)
            lr_file = os.path.join(self.cfg.output_folder, user_lr)
            lr_table = np.genfromtxt(lr_file)

            if len(lr_table.shape) == 1:
                lr_table = lr_table.reshape((1, ) + lr_table.shape)

            if len(lr_table.shape) != 2 or lr_table.shape[1] != 3:
                raise ValueError("Invalid lr.csv format")
            lr_table = [[tuple(lrt[0:2]), lrt[2]] for lrt in lr_table]
            lr_table.sort()

            lr = None
            # get lr from the table, one line before where the table is bigger
            for table_epoch in lr_table:
                if table_epoch[0] > tuple(epoch):
                    break
                else:
                    lr = table_epoch[1]
            if lr is None:
                raise ValueError(
                    "csv learning rate not specified for epoch {}".format(epoch))
            return lr

        try:
            # Float => Constant LR
            lr = float(user_lr)
            return lr
        except (ValueError, TypeError):
            pass

        try:
            # List => Exponentially decaying LR
            length = len(user_lr)
            lr_init = float(user_lr[0])
            lr_decay = float(user_lr[1])
            if length != 2:
                raise LookupError(
                    "{} (Your tuple has length {})".format(error_msg,
                                                           len(user_lr)))

            lr = lr_init * (1 - lr_decay) ** (
                        (epoch[1] - 1) + (epoch[0] - 1) * no_train_files)
            return lr
        except (ValueError, TypeError):
            pass

        try:
            # Callable => User defined function
            n_params = len(signature(user_lr).parameters)
            if n_params != 2:
                raise TypeError("A custom learning rate function must have two "
                                "input parameters: The epoch and the file number. "
                                "(yours has {})".format(n_params))
            lr = user_lr(epoch[0], epoch[1])
            return lr
        except (ValueError, TypeError):
            raise TypeError("{} (You gave {} of type {}) ".format(
                error_msg, user_lr, type(user_lr)))


def split_name_of_predfile(file):
    """
    Get epoch, fileno, cal fileno from the name of a predfile.

    Parameters
    ----------
    file : str
        Like this: model_epoch_XX_file_YY_on_USERLIST_val_file_ZZ.h5

    Returns
    -------
    epoch , file_no, val_file_no : tuple(int)
        As integers.

    """
    file_base = os.path.splitext(file)[0]
    rest, val_file_no = file_base.split("_val_file_")
    rest, file_no = rest.split("_on_")[0].split("_file_")
    epoch = rest.split("_epoch_")[-1]

    epoch, file_no, val_file_no = map(int, [epoch, file_no, val_file_no])

    return epoch, file_no, val_file_no


def h5_get_number_of_rows(h5_filepath, datasets=None):
    """
    Gets the total number of rows of of a .h5 file.

    Multiple dataset names can be given as a list to check if they all
    have the same number of rows (axis 0).

    Parameters
    ----------
    h5_filepath : str
        filepath of the .h5 file.
    datasets : list
        Optional, The names of datasets in the file to check.

    Returns
    -------
    number_of_rows: int
        number of rows of the .h5 file in the first dataset.

    Raises
    ------
    AssertionError
        If the given datasets do not have the same no of rows.

    """
    with h5py.File(h5_filepath, 'r') as f:
        if datasets is None:
            datasets = [x for x in list(f.keys())]

        number_of_rows = [f[dataset].shape[0] for dataset in datasets]
    if not number_of_rows.count(number_of_rows[0]) == len(number_of_rows):
        err_msg = "Datasets do not have the same number of samples " \
                  "in file " + h5_filepath
        for i, dataset in enumerate(datasets):
            err_msg += "\nDataset: {}\tSamples: {}".format(dataset,
                                                           number_of_rows[i])
        raise AssertionError(err_msg)
    return number_of_rows[0]


def use_local_tmpdir(files):
    """
    Copies given files to the local temp folder.

    Parameters
    ----------
    files : dict
        Dict containing the file pathes.

    Returns
    -------
    files_ssd : dict
        Dict with updated SSD/scratch filepaths.

    """
    local_scratch_path = os.environ['TMPDIR']
    files_ssd = {}

    for input_key in files:
        old_pathes = files[input_key]
        new_pathes = []
        for f_path in old_pathes:
            # copy to /scratch node-local SSD
            f_path_ssd = os.path.join(local_scratch_path,
                                      os.path.basename(f_path))
            print("Copying", f_path, "\nto", f_path_ssd)
            shutil.copy2(f_path, local_scratch_path)
            new_pathes.append(f_path_ssd)
        files_ssd[input_key] = tuple(new_pathes)

    print('Finished copying to local tmpdir folder.')
    return files_ssd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Run a test of the entire code on dummy data. """
import tempfile
import numpy as np
import h5py
import os
import sys
import shutil

import tensorflow as tf
import orcanet
from orcanet.core import Organizer
from orcanet.model_builder import ModelBuilder
from orcanet_contrib.custom_objects import get_custom_objects


class TestIntegration(tf.test.TestCase):
    """ Run the actual training to see if it throws any errors. """
    def setUp(self):
        """
        Make a .temp directory in the current working directory, generate
        dummy data in it and set up the cfg object.

        """
        # Pathes
        # Temporary output folder
        self.tdir = tempfile.TemporaryDirectory()
        self.temp_dir = self.tdir.name
        self.output_folder = os.path.join(self.temp_dir, "model")

        # Pathes to temp dummy data that will get generated
        train_inp = (self.temp_dir + "/train1.h5", self.temp_dir + "/train2.h5")
        self.train_pathes = {"testing_input": train_inp}

        val_inp = (self.temp_dir + "/val1.h5", self.temp_dir + "/val2.h5")
        self.val_pathes = {"testing_input": val_inp}

        # The config file to load
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        config_file = os.path.join(self.data_folder, "config_test.toml")

        # Make dummy data of given shape
        self.shape = (3, 3, 3, 3)
        for path1, path2 in (train_inp, val_inp):
            make_dummy_data(path1, path2, self.shape)

        def make_orga():
            orga = Organizer(self.output_folder, config_file=config_file)
            orga.cfg._files_dict["train"] = self.train_pathes
            orga.cfg._files_dict["val"] = self.val_pathes
            orga.cfg._list_file = "test.toml"
            orga.cfg.zero_center_folder = self.temp_dir
            orga.cfg.label_modifier = label_modifier
            orga.cfg.custom_objects = get_custom_objects()
            return orga

        self.make_orga = make_orga

    def tearDown(self):
        """ Remove the .temp directory. """
        self.tdir.cleanup()

    def test_integration_zero_center(self):
        """ Calculate the zero center image and check if it works properly. """
        orga = self.make_orga()
        xs_mean = orga.get_xs_mean()
        target_xs_mean = np.ones(self.shape)/4
        self.assertTrue(np.allclose(xs_mean["testing_input"], target_xs_mean))

        file = orga.cfg.zero_center_folder + "/" + orga.cfg._list_file + \
            '_input_' + "testing_input" + '.npz'
        zero_center_used_ip_files = np.load(file)['zero_center_used_ip_files']
        self.assertTrue(np.array_equal(zero_center_used_ip_files,
                                       orga.cfg._files_dict["train"]["testing_input"]))

    def test_integration_multi_input_model(self):
        """
        Run whole script on dummy data to see if it throws an error.

        Build a model with the ModelBuilder.
        Train for 2 epochs.
        Reset organizer.
        Resume for 1 epoch with different lr and sample modifier.
        Use orga.train and validate once each.
        Predict.

        """
        orga = self.make_orga()

        model_file = os.path.join(self.data_folder, "model_test.toml")
        builder = ModelBuilder(model_file)
        initial_model = builder.build(orga)

        orga.train_and_validate(initial_model, epochs=2)

        def test_learning_rate(epoch, fileno):
            lr = 0.001 * (epoch + 0.1*fileno)
            return lr

        def test_modifier(info_blob):
            xs = info_blob["x_values"]
            xs = {key: xs[key] for key in xs}
            return xs

        orga = self.make_orga()
        orga.cfg.learning_rate = test_learning_rate
        orga.cfg.sample_modifier = test_modifier
        orga.train_and_validate(epochs=1)
        orga.train()
        orga.validate()
        orga.predict()


class TestFullExample(tf.test.TestCase):
    """
    Test the full example in the examples folder, as it is used on the docs.
    Must go thorugh without any errors.
    """
    def test_full_example(self):
        init_dir = os.getcwd()

        example_dir = os.path.join(
            os.path.dirname(orcanet.__path__[0]), "examples/full_example")
        os.chdir(example_dir)
        sys.path.append(example_dir)
        output_dir = os.path.join(example_dir, "output")
        if os.path.exists(output_dir):
            raise FileExistsError(output_dir)
        try:
            import full_example
        finally:
            if os.path.exists(output_dir):
                shutil.rmtree(output_dir, ignore_errors=True)
            os.chdir(init_dir)


def make_dummy_data(filepath1, filepath2, shape):
    """
    Make a total of 100 ones vs 300 zeroes of dummy data over two files.

    Parameters
    ----------
    filepath1 : str
        Path to file 1.
    filepath2 : str
        Path to file 2.
    shape : tuple
        Shape of the data, not including sample dimension.

    """
    xs1 = np.concatenate([np.ones((75,) + shape), np.zeros((75,) + shape)])
    xs2 = np.concatenate([np.ones((25,) + shape), np.zeros((225,) + shape)])

    dtypes = [('dir_x', '<f8'), ]
    ys1 = np.ones((150, len(dtypes))).ravel().view(dtype=dtypes)
    ys2 = np.ones((250, len(dtypes))).ravel().view(dtype=dtypes)

    for xs, ys, filepath in ((xs1, ys1, filepath1), (xs2, ys2, filepath2)):
        h5f = h5py.File(filepath, 'w')
        h5f.create_dataset('x', data=xs, dtype='uint8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)
        h5f.close()


def label_modifier(info_blob):
    y_values = info_blob["y_values"]
    ys = dict()

    ys['dx'], ys['dx_err'] = y_values['dir_x'], y_values['dir_x']

    for key_label in ys:
        ys[key_label] = ys[key_label].astype(np.float32)
    return ys



import numpy as np
import os
from orcanet.utilities.visualization import plot_history
from orcanet.in_out import get_subfolder


class HistoryHandler:
    """
    For reading and plotting data from summary and train log files.

    """
    def __init__(self, main_folder):
        self.main_folder = main_folder
        self.summary_filename = "summary.txt"

    @property
    def summary_file(self):
        main_folder = self.main_folder
        if not main_folder[-1] == "/":
            main_folder += "/"
        return main_folder + self.summary_filename

    @property
    def train_log_folder(self):
        return get_subfolder(self.main_folder, "train_log")

    def plot_metric(self, metric_name, **kwargs):
        """
        Plot the training and validation history of a metric.

        This will read out data from the summary file, as well as
        all training log files, and plot them over the epoch.

        Parameters
        ----------
        metric_name : str
            Name of the metric to be plotted over the epoch. This name is what
            was written in the head line of the summary.txt file, except without
            the train_ or val_ prefix.
        kwargs
            Keyword arguments for the plot_history function.

        """
        summary_data = self.get_summary_data()
        full_train_data = self.get_train_data()
        summary_label = "val_" + metric_name

        if metric_name not in full_train_data.dtype.names:
            raise ValueError(
                "Train log metric name {} unknown, must be one of {}".format(
                    metric_name, self.get_metrics()))
        if summary_label not in summary_data.dtype.names:
            raise ValueError(
                "Summary metric name {} unknown, must be one of {}".format(
                    summary_label, self.get_metrics()))

        if summary_data["Epoch"].shape == (0,):
            # When no lines are present in the summary.txt file.
            val_data = None
        else:
            val_data = [summary_data["Epoch"], summary_data[summary_label]]

        if full_train_data["Batch_float"].shape == (0,):
            # When no lines are present
            raise ValueError("Can not make summary plot: Training log files "
                             "contain no data!")
        else:
            train_data = [full_train_data["Batch_float"],
                          full_train_data[metric_name]]

        # if no validation has been done yet
        if np.all(np.isnan(val_data)[1]):
            val_data = None

        if "y_label" not in kwargs:
            kwargs["y_label"] = metric_name

        plot_history(train_data, val_data, **kwargs)

    def plot_lr(self, **kwargs):
        """
        Plot the learning rate over the epochs.

        Parameters
        ----------
        kwargs
            Keyword arguments for the plot_history function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The plot.

        """
        summary_data = self.get_summary_data()

        epoch = summary_data["Epoch"]
        lr = summary_data["LR"]
        # plot learning rate like val data (connected dots)
        val_data = (epoch, lr)

        if "y_label" not in kwargs:
            kwargs["y_label"] = "Learning rate"
        if "legend" not in kwargs:
            kwargs["legend"] = False

        plot_history(train_data=None, val_data=val_data, logy=True,
                     y_lims=None, **kwargs)

    def get_metrics(self):
        """
        Get the name of the metrics from the first line in the summary file.

        This will be the actual name of the metric, i.e. "loss" and not
        "train_loss" or "val_loss".

        Returns
        -------
        all_metrics : List
            A list of the metrics.

        """
        summary_data = self.get_summary_data()
        all_metrics = []
        for keyword in summary_data.dtype.names:
            if keyword == "Epoch" or keyword == "LR":
                continue
            if "train_" in keyword:
                keyword = keyword.split("train_")[-1]
            else:
                keyword = keyword.split("val_")[-1]
            if keyword not in all_metrics:
                all_metrics.append(keyword)
        return all_metrics

    def get_summary_data(self):
        """
        Read out the summary file in the output folder.

        Returns
        -------
        summary_data : ndarray
            Numpy structured array with the column names as datatypes.
            Its shape is the number of lines with data.

        """
        summary_data = self._load_txt(self.summary_file)
        if summary_data.shape == ():
            # When only one line is present
            summary_data = summary_data.reshape(1,)
        return summary_data

    def get_best_epoch_info(self, metric="val_loss", mini=True):
        """
        Get the line in the summary file where the given metric is best, i.e.
        either minimal  or maximal.

        Parameters
        ----------
        metric : str
            Which metric to look up.
        mini : bool
            If true, look up the minimum. Else the maximum.

        Raises
        ------
        ValueError
            If there is no best line (e.g. no validation has been done).

        """
        summary_data = self.get_summary_data()
        metric_data = summary_data[metric]

        if all(np.isnan(metric_data)):
            raise ValueError("Can not find best epoch in summary.txt")

        if mini:
            opt_loss = np.nanmin(metric_data)
        else:
            opt_loss = np.nanmax(metric_data)

        best_index = np.where(metric_data == opt_loss)[0]
        # if multiple epochs with same loss, take first
        best_index = min(best_index)
        best_line = summary_data[best_index]
        return best_line

    def get_best_epoch_fileno(self, metric="val_loss", mini=True):
        """
        Get the epoch/fileno tuple where the given metric is smallest.
        """
        best_line = self.get_best_epoch_info(metric=metric, mini=mini)
        best_epoch_float = best_line["Epoch"]
        epoch, fileno = self._transform_epoch(best_epoch_float)
        return epoch, fileno

    def _transform_epoch(self, epoch_float):
        """
        Transfrom the epoch_float read from a file to a tuple.

        (By just counting the number of lines in the given epoch).
        TODO Hacky, (epoch, filno) should probably be written to the summary.
        """
        summary_data = self.get_summary_data()

        epoch = int(np.floor(epoch_float - 1e-8))
        # all lines in the epoch of epoch_float
        indices = np.where(np.floor(summary_data["Epoch"] - 1e-8) == epoch)[0]
        lines = summary_data[indices]
        fileno = int(np.where(lines["Epoch"] == epoch_float)[0]) + 1
        epoch += 1
        return epoch, fileno

    def get_column_names(self):
        """
        Get the str in the first line in each column.

        Returns
        -------
        tuple : column_names
            The names in the same order as they appear in the summary.txt.

        """
        summary_data = self.get_summary_data()
        column_names = summary_data.dtype.names
        return column_names

    def get_train_data(self):
        """
        Read out all training logfiles in the output folder.

        Read out the data from the summary.txt file, and from all training
        log files in the train_log folder, which is in the same directory
        as the summary.txt file.

        Returns
        -------
        summary_data : numpy.ndarray
            Structured array containing the data from the summary.txt file.
            Its shape is the number of lines with data.

        """
        # list of all files in the train_log folder of this model
        files = os.listdir(self.train_log_folder)
        train_file_data = []
        for file in files:
            if not (file.startswith("log_epoch_") and file.endswith(".txt")):
                continue
            filepath = os.path.join(self.train_log_folder, file)
            if os.path.getsize(filepath) == 0:
                continue

            # file is sth like "log_epoch_1_file_2.txt", extract epoch & fileno:
            epoch, file_no = [int(file.split(".")[0].split("_")[i]) for i in [2, 4]]
            file_data = self._load_txt(filepath)
            train_file_data.append([[epoch, file_no], file_data])

        if len(train_file_data) == 0:
            raise OSError(f"No train files found in {self.train_log_folder}!")

        # sort so that earlier epochs come first
        train_file_data.sort()
        full_train_data = train_file_data[0][1]
        for [epoch, file_no], file_data in train_file_data[1:]:
            full_train_data = np.append(full_train_data, file_data)

        if full_train_data.shape == ():
            # When only one line is present
            full_train_data = full_train_data.reshape(1,)
        return full_train_data

    def get_state(self):
        """
        Get the state of a training.

        For every line in the summary logfile, get a dict with the epoch
        as a float, and is_trained and is_validated bools.

        Returns
        -------
        state_dicts : List
            List of dicts.

        """
        summary_data = self.get_summary_data()
        state_dicts = []
        names = summary_data.dtype.names

        for line in summary_data:
            val_losses, train_losses = {}, {}
            for name in names:
                if name.startswith("val_"):
                    val_losses[name] = line[name]
                elif name.startswith("train_"):
                    train_losses[name] = line[name]
                elif name not in ["Epoch", "LR"]:
                    raise NameError(
                        "Invalid summary file: Invalid column name {}: must be "
                        "either Epoch, LR, or start with val_ or train_".format(name))
            # if theres any not-nan entry, consider it completed
            is_trained = any(~np.isnan(tuple(train_losses.values())))
            is_validated = any(~np.isnan(tuple(val_losses.values())))

            state_dicts.append({
                "epoch": line["Epoch"],
                "is_trained": is_trained,
                "is_validated": is_validated, })

        return state_dicts

    @staticmethod
    def _load_txt(filepath):
        # TODO suboptimal that n/a gets replaced by np.nan, because this
        #  means that legitamte, not availble cells can not be distinguished
        #  from failed 'nan' metric values produced by training.
        file_data = np.genfromtxt(
            filepath,
            names=True,
            delimiter="|",
            autostrip=True,
            comments="--",
            missing_values="n/a",
            filling_values=np.nan
        )
        # replace inf with nan so it can be plotted
        for column_name in file_data.dtype.names:
            x = file_data[column_name]
            x[np.isinf(x)] = np.nan
        return file_data



import tempfile
from unittest import TestCase
from unittest.mock import MagicMock
import os
import warnings
import h5py
import numpy as np
import tensorflow as tf
import tensorflow.keras as ks
import tensorflow.keras.layers as layers

from orcanet.core import Organizer
from orcanet.backend import train_model, validate_model, make_model_prediction, weighted_average
import orcanet.lib as lib


class TestFunctions(TestCase):
    def test_weighted_average(self):
        # metrics [A, B, C]
        histories = [
            [0, 1, 2],  # file 1
            [3, 4, 5],  # file 2
        ]

        file_sizes = [
            1,
            3,
        ]

        target = [
            (0*1 + 3*3)/4, (1*1 + 4*3)/4, (2*1 + 5*3)/4,
        ]
        averaged_histories = weighted_average(histories, file_sizes)

        self.assertSequenceEqual(averaged_histories, target)

    def test_weighted_average_one_file(self):
        # metrics [A, B, C]
        histories = [
            [0, 1, 2],  # file 1
        ]

        file_sizes = [
            1,
        ]

        target = [
            0, 1, 2
        ]
        averaged_histories = weighted_average(histories, file_sizes)

        self.assertSequenceEqual(averaged_histories, target)

    def test_weighted_average_one_metric(self):
        # metrics [A, ]
        histories = [
            [0, ],  # file 1
            [3, ],  # file 2
        ]

        file_sizes = [
            1,
            3,
        ]

        target = [
            (0*1 + 3*3)/4,
        ]
        averaged_histories = weighted_average(histories, file_sizes)

        self.assertSequenceEqual(averaged_histories, target)

    def test_weighted_average_one_metric_one_file(self):
        # metrics [A, ]
        histories = [
            [1, ],  # file 1
        ]

        file_sizes = [
            1,
        ]

        target = [
            1,
        ]
        averaged_histories = weighted_average(histories, file_sizes)

        self.assertSequenceEqual(averaged_histories, target)


class TestTrainValidatePredict(tf.test.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.tdir = tempfile.TemporaryDirectory()
        cls.temp_dir = cls.tdir.name
        cls.pred_dir = os.path.join(cls.temp_dir, "predictions")

        os.mkdir(cls.pred_dir)

        cls.pred_filepath = os.path.join(
            cls.pred_dir,
            'pred_model_epoch_1_file_3_on_listfilename_val_file_1.h5')
        cls.file_sizes = [500, ]
        # make some dummy data
        cls.inp_A_file = {
            "path": os.path.join(cls.temp_dir, "input_A_file.h5"),
            "shape": (2, 3),
            "size": cls.file_sizes[0],
        }

        cls.inp_B_file = {
            "path": os.path.join(cls.temp_dir, "input_B_file.h5"),
            "shape": (3, 4),
            "size": cls.file_sizes[0],
        }

        cls.filepaths = {
            "input_A": (cls.inp_A_file["path"],),
            "input_B": (cls.inp_B_file["path"],),
        }

        cls.input_shapes = {
            "input_A": cls.inp_A_file["shape"],
            "input_B": cls.inp_B_file["shape"],
        }

        cls.output_shapes = {
            "mc_A": 1,
            "mc_B": 1,
        }

        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.inp_A_file, mode="half")
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.inp_B_file, mode="half")

    def setUp(self):
        self.orga = Organizer(self.temp_dir)
        self.orga.cfg.batchsize = 9

        self.orga.io.get_local_files = MagicMock(return_value=self.filepaths)
        self.orga.io.get_file_sizes = MagicMock(return_value=self.file_sizes)
        self.orga.cfg.get_list_file = MagicMock(return_value='/path/to/a/listfilename.toml')
        self.orga.io.get_next_pred_path = MagicMock(return_value=self.pred_filepath)

        self.model = build_dummy_model(self.input_shapes, self.output_shapes)
        self.model.compile(loss="mse", optimizer="sgd")
        self.orga._auto_label_modifier = lib.label_modifiers.ColumnLabels(self.model)

    @classmethod
    def tearDownClass(cls):
        cls.tdir.cleanup()

    def test_train(self):
        epoch = (1, 1)
        batch_nos = []
        batch_print_callback = ks.callbacks.LambdaCallback(
            on_batch_begin=lambda batch, logs: batch_nos.append(batch))
        self.orga.cfg.callback_train = batch_print_callback

        history = train_model(self.orga, self.model, epoch, batch_logger=False)
        target = {  # TODO these sometimes change, so I set rtol to be high. But WHY???
            'loss': 18.252519607543945,
            'mc_A_loss': 9.647336959838867,
            'mc_B_loss': 8.605181694030762,
        }
        print(history, target)
        assert_dict_arrays_equal(history, target, rtol=0.8)
        self.assertSequenceEqual(batch_nos, list(range(int(self.file_sizes[0]/self.orga.cfg.batchsize))))

    def test_validate(self):
        history = validate_model(self.orga, self.model)
        # input to model is ones
        # --> after concatenate: 18 ones
        # Output of each output layer = 18
        # labels: 0 and 1
        # --> loss (18-0)^2 and (18-1)^2
        target = {
            'loss': 18**2 + 17**2,
            'mc_A_loss': 18**2,
            'mc_B_loss': 17**2,
        }
        self.model.summary()
        self.assertDictEqual(history, target)

    def test_predict(self):
        # dummy values
        epoch, fileno = 1, 3
        # mock get_latest_prediction_file_no
        # self.orga.io.get_latest_prediction_file_no = MagicMock(return_value=None)

        try:
            make_model_prediction(self.orga, self.model, epoch, fileno)

            file_cntn = {}
            with h5py.File(self.pred_filepath, 'r') as file:
                for key in file.keys():
                    file_cntn[key] = np.array(file[key])
        finally:
            os.remove(self.pred_filepath)

        target_datasets = [
            'label_mc_A', 'label_mc_B', 'pred_mc_A', 'pred_mc_B', 'y_values'
        ]
        target_shapes = [
            (500,), (500,), (500, 1), (500, 1), (500,)
        ]
        target_contents = [
            np.zeros(target_shapes[0]),
            np.ones(target_shapes[1]),
            np.ones(target_shapes[2]) * 18,
            np.ones(target_shapes[3]) * 18,
            self.train_A_file_1_ctnt[1],
        ]
        shapes_dict = dict(zip(target_datasets, target_shapes))
        contents_dict = dict(zip(target_datasets, target_contents))

        target_mc_names = ('mc_A', 'mc_B')

        datasets = list(file_cntn.keys())
        shapes = [file_cntn[key].shape for key in datasets]
        mc_dtype_names = file_cntn["y_values"].dtype.names

        self.assertSequenceEqual(datasets, target_datasets)
        self.assertSequenceEqual(shapes, target_shapes)
        self.assertSequenceEqual(mc_dtype_names, target_mc_names)
        for i, key in enumerate(target_datasets):
            value = file_cntn[key]
            target = contents_dict[key]
            np.testing.assert_array_equal(value, target)


def save_dummy_h5py(path, shape, size, mode="asc"):
    """
    if mode == asc:
        xs[0] is np.zeros, xs[1] is np.ones, etc
        ys[0] is ndarray with 0.5s, ys[1] with 1.5s etc

    if mode == half:
        xs is ones,
        ys[:, 0] is 0, ys[:, 1] is 1

    """
    xs = np.ones((size,) + shape)

    dtypes = [('mc_A', '<f8'), ('mc_B', '<f8'), ]
    ys = np.ones((size, 2))

    if mode == "asc":
        for sample_no in range(len(xs)):
            xs[sample_no, ...] = sample_no
            ys[sample_no, ...] = sample_no + 0.5

    elif mode == "half":
        ys[:, 0] = 0.

    else:
        raise AssertionError

    ys = ys.ravel().view(dtype=dtypes)

    with h5py.File(path, 'w') as h5f:
        h5f.create_dataset('x', data=xs, dtype='<f8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)
    return xs, ys


def assert_dict_arrays_equal(a, b, rtol=1e-4):
    for key, value in a.items():
        assert key in b
        np.testing.assert_allclose(a[key], b[key], rtol=rtol)
        if not np.array_equal(a[key], b[key]):
            warnings.warn(f"Arrays not equal:{a[key]} != {b[key]}")


def assert_equal_struc_array(a, b):
    """  np.testing.assert_array_equal does not work for arrays containing nans...
     so test individual instead. """
    np.testing.assert_array_equal(a.dtype, b.dtype)
    for name in a.dtype.names:
        np.testing.assert_almost_equal(a[name], b[name])


def build_dummy_model(input_shapes, output_shapes):
    """

    Parameters
    ----------
    input_shapes : dict
    output_shapes : dict

    Returns
    -------
    model : keras model

    """
    inputs, flattend = [], []
    for name, shape in input_shapes.items():
        inp = layers.Input(shape, name=name)
        flat = layers.Flatten()(inp)
        inputs.append(inp)
        flattend.append(flat)

    conc = layers.Concatenate()(flattend)

    outputs = []
    for name, shape in output_shapes.items():
        outputs.append(layers.Dense(
            shape, name=name, kernel_initializer="Ones",
            bias_initializer="Zeros")(conc))

    model = ks.models.Model(inputs, outputs)
    return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for training and validating NN's, as well as evaluating them.
"""
import time
import h5py
import numpy as np
import os

import orcanet
from orcanet.logging import BatchLogger
import orcanet.utilities.nn_utilities as nn_utilities
from orcanet.in_out import h5_get_number_of_rows
import orcanet.h5_generator as h5_generator
import orcanet.lib.dataset_modifiers as dataset_modifiers


def train_model(orga, model, epoch, batch_logger=False):
    """
    Train a model on one file and return the history.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras.Model
        A compiled keras model.
    epoch : tuple
        Current epoch and the no of the file to train on.
    batch_logger : bool
        Use the orcanet batchlogger to log the training.

    Returns
    -------
    history : dict
        The history of the training on this file. A record of training
        loss values and metrics values.

    """
    callbacks = [
        nn_utilities.RaiseOnNaN(),
        nn_utilities.TimeModel(print_func=orga.io.print_log),
    ]
    if batch_logger:
        callbacks.append(BatchLogger(orga, epoch))
    if orga.cfg.callback_train is not None:
        try:
            callbacks.extend(orga.cfg.callback_train)
        except TypeError:
            callbacks.append(orga.cfg.callback_train)

    training_generator = h5_generator.get_h5_generator(
        orga,
        files_dict=orga.io.get_file("train", epoch[1]),
        f_size=orga.cfg.n_events,
        phase="training",
        zero_center=orga.cfg.zero_center_folder is not None,
        shuffle=orga.cfg.shuffle_train,
    )
    # status tf.2.5: In order to use ragged Tensors as input to fit,
    #  we have to use a tf dataset and not a generator
    dataset = h5_generator.make_dataset(training_generator)

    history = model.fit(
        dataset,
        steps_per_epoch=len(training_generator),
        verbose=orga.cfg.verbose_train,
        max_queue_size=orga.cfg.max_queue_size,
        callbacks=callbacks,
        initial_epoch=epoch[0] - 1,
        epochs=epoch[0],
    )
    training_generator.print_timestats(print_func=orga.io.print_log)
    # get a dict with losses and metrics
    # only trained for one epoch, so value is list of len 1
    history = {key: value[0] for key, value in history.history.items()}
    return history


def validate_model(orga, model):
    """
    Validates a model on all validation files and return the history.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras.Model
        A compiled keras model.

    Returns
    -------
    history : dict
        The history of the validation on all files. A record of validation
        loss values and metrics values.

    """
    # One history for each val file
    histories = []
    f_sizes = orga.io.get_file_sizes("val")

    for i, files_dict in enumerate(orga.io.yield_files("val")):
        val_generator = h5_generator.get_h5_generator(
            orga,
            files_dict,
            f_size=orga.cfg.n_events,
            phase="validation",
            zero_center=orga.cfg.zero_center_folder is not None,
        )
        # status tf.2.5: In order to use ragged Tensors as input to fit,
        #  we have to use a tf dataset and not a generator
        dataset = h5_generator.make_dataset(val_generator)
        history_file = model.evaluate(
            dataset,
            steps=len(val_generator),
            max_queue_size=orga.cfg.max_queue_size,
            verbose=orga.cfg.verbose_val)
        if not isinstance(history_file, list):
            history_file = [history_file, ]
        histories.append(history_file)

    # average over all val files
    history = weighted_average(histories, f_sizes)
    # This history is just a list, not a dict like with fit_generator
    # so transform to dict
    history = dict(zip(model.metrics_names, history))

    return history


def weighted_average(histories, f_sizes):
    """
    Average multiple histories, weighted with the file size.

    Each history can have multiple metrics, which are averaged seperatly.

    Parameters
    ----------
    histories : List
        List of histories, one for each file. Each history is also
        a list: each entry is a different loss or metric.
    f_sizes : List
        List of the file sizes, in the same order as the histories, i.e.
        the file of histories[0] has the length f_sizes[0].

    Returns
    -------
    wgtd_average : List
        The weighted averaged history. Has the same length as each
        history in the histories List, i.e. one entry per loss or metric.

    """
    assert len(histories) == len(f_sizes)
    rltv_fsizes = [f_size/sum(f_sizes) for f_size in f_sizes]
    wgtd_average = np.dot(np.transpose(histories), rltv_fsizes)

    return wgtd_average.tolist()


def h5_inference(orga, model, files_dict, output_path, samples=None, use_def_label=True):
    """
    Let a model predict on all samples in a h5 file, and save it as a h5 file.

    Per default, the h5 file will contain a datagroup y_values straight from
    the given files, as well as two datagroups per output layer of the network,
    which have the labels and the predicted values in them as numpy arrays,
    respectively.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras.Model
        Trained Keras model of a neural network.
    files_dict : dict
        Dict mapping model input names to h5 file paths.
    output_path : str
        Name of the output h5 file containing the predictions.
    samples : int, optional
        Dont use all events in the file, but instead only the given number.
    use_def_label : bool
        If True and no label modifier is given by user, use the default
        label modifier instead of none.

    """
    file_size = h5_get_number_of_rows(
        list(files_dict.values())[0],
        datasets=[orga.cfg.key_y_values])
    generator = h5_generator.get_h5_generator(
        orga,
        files_dict,
        zero_center=orga.cfg.zero_center_folder is not None,
        keras_mode=False,
        use_def_label=use_def_label,
        phase="inference",
    )
    itergen = iter(generator)

    if samples is None:
        steps = len(generator)
    else:
        steps = int(samples / orga.cfg.batchsize)
    print_every = max(100, min(int(round(steps/10, -2)), 1000))
    model_time_total = 0.
    dataset_last_element = {}

    temp_output_path = os.path.join(
        os.path.dirname(output_path),
        "temp_" + os.path.basename(output_path) + "_" +
        time.strftime("%d-%m-%Y-%H-%M-%S", time.gmtime()))
    print(f"Creating temporary file {temp_output_path}")
    with h5py.File(temp_output_path, 'x') as h5_file:
        # add version and paths of h5files
        h5_file.attrs.create("orcanet", orcanet.__version__)
        for input_key, file in files_dict.items():
            h5_file.attrs.create(f"orcanet_inp_{input_key}", file)

        for s in range(steps):
            if s % print_every == 0:
                print('Predicting in step {}/{} ({:0.2%})'.format(
                    s, steps, s/steps))

            info_blob = next(itergen)

            start_time = time.time()
            y_pred = model.predict_on_batch(info_blob["xs"])
            model_time_total += time.time() - start_time

            if not isinstance(y_pred, list):
                # if only one output, transform to a list
                y_pred = [y_pred]
            # transform y_pred to dict
            y_pred = {
                out: y_pred[i] for i, out in enumerate(model.output_names)}
            info_blob["y_pred"] = y_pred

            if info_blob.get("org_batchsize") is not None:
                _slice_to_size(info_blob)

            if orga.cfg.dataset_modifier is None:
                datasets = dataset_modifiers.as_array(info_blob)
            else:
                datasets = orga.cfg.dataset_modifier(info_blob)

            for dataset_name, data in datasets.items():
                if s == 0:
                    # create datasets in the first step
                    h5_file.create_dataset(
                        dataset_name,
                        shape=(file_size,) + data.shape[1:],
                        dtype=data.dtype,
                        chunks=True,
                        compression="gzip",
                        compression_opts=1,
                        shuffle=True,
                    )
                    dataset_last_element[dataset_name] = 0

                ix = dataset_last_element[dataset_name]
                h5_file[dataset_name][ix:(ix + data.shape[0])] = data
                dataset_last_element[dataset_name] += data.shape[0]

    if os.path.exists(output_path):
        raise FileExistsError(
            f"{output_path} exists already! But file {temp_output_path} "
            f"is finished and can be safely used.")
    os.rename(temp_output_path, output_path)
    orga.io.print_log(f"\nPrediction completed on file {os.path.basename(output_path)}")
    orga.io.print_log("Statistics of model prediction:")
    orga.io.print_log(f"\tTotal time:\t{model_time_total / 60:.2f} min")
    orga.io.print_log(f"\tPer batch:\t{1000 * model_time_total / steps:.5} ms")
    generator.print_timestats(print_func=orga.io.print_log)


def _slice_to_size(info_blob):
    org_batchsize = info_blob["org_batchsize"]
    for input_key, x in info_blob["xs"].items():
        info_blob["xs"][input_key] = x[:org_batchsize]
    for output_key, y_pred in info_blob["y_pred"].items():
        info_blob["y_pred"][output_key] = y_pred[:org_batchsize]
    if info_blob.get("ys") is not None:
        for output_key, y in info_blob["ys"].items():
            info_blob["ys"][output_key] = y[:org_batchsize]


def make_model_prediction(orga, model, epoch, fileno, samples=None):
    """
    Let a model predict on all validation samples, and save it as a h5 file.

    Per default, the h5 file will contain a datagroup y_values straight from
    the given files, as well as two datagroups per output layer of the network,
    which have the labels and the predicted values in them as numpy arrays,
    respectively.

    Parameters
    ----------
    orga : orcanet.core.Organizer
        Contains all the configurable options in the OrcaNet scripts.
    model : keras.Model
        A compiled keras model.
    epoch : int
        Epoch of the last model training step in the epoch, file_no tuple.
    fileno : int
        File number of the last model training step in the epoch, file_no tuple.
    samples : int or None
        Number of events that should be predicted.
        If samples=None, the whole file will be used.

    """
    latest_pred_file_no = orga.io.get_latest_prediction_file_no(epoch, fileno)
    if latest_pred_file_no is None:
        latest_pred_file_no = 0

    # For every val file set (one set can have multiple files if
    # the model has multiple inputs):
    for f_number, files_dict in enumerate(orga.io.yield_files("val"), 1):
        if f_number <= latest_pred_file_no:
            continue

        pred_filepath = orga.io.get_pred_path(epoch, fileno, f_number)
        h5_inference(orga, model, files_dict, pred_filepath, samples=samples)



from unittest import TestCase
from unittest.mock import MagicMock
import os
import h5py
import numpy as np
import shutil
from pathlib import Path
from tensorflow.keras.models import Model
import tensorflow.keras.layers as layers

from orcanet.core import Configuration
from orcanet.in_out import IOHandler, split_name_of_predfile


class TestIOHandler(TestCase):
    @classmethod
    def setUpClass(cls):
        # super(TestIOHandler, cls).setUpClass()
        cls.temp_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                    "test_in_out")
        cls.pred_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                    "test_in_out", "predictions")
        os.mkdir(cls.temp_dir)
        os.mkdir(cls.pred_dir)

        # make dummy pred files
        Path(cls.pred_dir + '/pred_model_epoch_2_file_2_on_listname_val_file_1.h5').touch()
        Path(cls.pred_dir + '/pred_model_epoch_2_file_2_on_listname_val_file_2.h5').touch()

        cls.pred_filepaths = [cls.pred_dir + '/pred_model_epoch_2_file_2_on_listname_val_file_1.h5',
                              cls.pred_dir + '/pred_model_epoch_2_file_2_on_listname_val_file_2.h5']

        cls.init_dir = os.getcwd()
        os.chdir(cls.temp_dir)
        # make some dummy data
        cls.n_bins = {'input_A': (2, 3), 'input_B': (2, 3)}
        cls.train_sizes = [30, 50]
        cls.val_sizes = [40, 60]
        cls.file_names = (
            "/input_A_train_1.h5",
            "/input_A_train_2.h5",
            "/input_B_train_1.h5",
            "/input_B_train_2.h5",
            "/input_A_val_1.h5",
            "/input_A_val_2.h5",
            "/input_B_val_1.h5",
            "/input_B_val_2.h5",
        )
        cls.train_A_file_1 = {
            "path": cls.temp_dir + cls.file_names[0],
            "shape": cls.n_bins["input_A"],
            "value_xs": 1.1,
            "value_ys": 1.2,
            "size": cls.train_sizes[0],
        }
        cls.train_A_file_2 = {
            "path": cls.temp_dir + cls.file_names[1],
            "shape": cls.n_bins["input_A"],
            "value_xs": 1.3,
            "value_ys": 1.4,
            "size": cls.train_sizes[1],
        }
        cls.train_B_file_1 = {
            "path": cls.temp_dir + cls.file_names[2],
            "shape": cls.n_bins["input_B"],
            "value_xs": 2.1,
            "value_ys": 2.2,
            "size": cls.train_sizes[0],
        }
        cls.train_B_file_2 = {
            "path": cls.temp_dir + cls.file_names[3],
            "shape": cls.n_bins["input_B"],
            "value_xs": 2.3,
            "value_ys": 2.4,
            "size": cls.train_sizes[1],
        }
        cls.val_A_file_1 = {
            "path": cls.temp_dir + cls.file_names[4],
            "shape": cls.n_bins["input_A"],
            "value_xs": 3.1,
            "value_ys": 3.2,
            "size": cls.val_sizes[0],
        }
        cls.val_A_file_2 = {
            "path": cls.temp_dir + cls.file_names[5],
            "shape": cls.n_bins["input_A"],
            "value_xs": 3.1,
            "value_ys": 3.2,
            "size": cls.val_sizes[0],
        }
        cls.val_B_file_1 = {
            "path": cls.temp_dir + cls.file_names[6],
            "shape": cls.n_bins["input_B"],
            "value_xs": 4.1,
            "value_ys": 4.2,
            "size": cls.val_sizes[1],
        }
        cls.val_B_file_2 = {
            "path": cls.temp_dir + cls.file_names[7],
            "shape": cls.n_bins["input_B"],
            "value_xs": 4.1,
            "value_ys": 4.2,
            "size": cls.val_sizes[1],
        }
        cls.train_A_file_1_ctnt = save_dummy_h5py(**cls.train_A_file_1)
        cls.train_A_file_2_ctnt = save_dummy_h5py(**cls.train_A_file_2)
        cls.train_B_file_1_ctnt = save_dummy_h5py(**cls.train_B_file_1)
        cls.train_B_file_2_ctnt = save_dummy_h5py(**cls.train_B_file_2)
        cls.val_A_file_1_ctnt = save_dummy_h5py(**cls.val_A_file_1)
        cls.val_A_file_2_ctnt = save_dummy_h5py(**cls.val_A_file_2)
        cls.val_B_file_1_ctnt = save_dummy_h5py(**cls.val_B_file_1)
        cls.val_B_file_2_ctnt = save_dummy_h5py(**cls.val_B_file_2)

    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.output_folder = self.data_folder + "/dummy_model"

        list_file = self.data_folder + "/in_out_test_list.toml"
        config_file = None

        cfg = Configuration(self.output_folder, list_file, config_file)
        self.batchsize = 3
        cfg.batchsize = self.batchsize
        self.io = IOHandler(cfg)

        # mock get_subfolder, but only in case of predictions argument
        original_get_subfolder = self.io.get_subfolder
        mocked_result = self.pred_dir

        def side_effect(key, create=False):
            if key == 'predictions':
                return mocked_result
            else:
                return original_get_subfolder(key, create)

        self.io.get_subfolder = MagicMock(side_effect=side_effect)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.train_A_file_1["path"])
        os.remove(cls.train_A_file_2["path"])
        os.remove(cls.train_B_file_1["path"])
        os.remove(cls.train_B_file_2["path"])
        os.remove(cls.val_A_file_1["path"])
        os.remove(cls.val_A_file_2["path"])
        os.remove(cls.val_B_file_1["path"])
        os.remove(cls.val_B_file_2["path"])

        os.chdir(cls.init_dir)
        shutil.rmtree(cls.temp_dir)

    def test_copy_to_ssd(self):
        self.io.cfg.use_scratch_ssd = True
        # make temporary directory
        temp_temp_dir = self.temp_dir + "/scratch"
        os.mkdir(temp_temp_dir)

        # change env variable TMPDIR to this dir (TMPDIR not defined in gitrunner)
        if "TMPDIR" in os.environ:
            tempdir_environ = os.environ["TMPDIR"]
        else:
            tempdir_environ = None

        try:
            os.environ["TMPDIR"] = temp_temp_dir
            scratch_dir = temp_temp_dir

            target_dirs_train = {
                 "input_A": (scratch_dir + self.file_names[0], scratch_dir + self.file_names[1]),
                 "input_B": (scratch_dir + self.file_names[2], scratch_dir + self.file_names[3]),
            }
            target_dirs_val = {
                 "input_A": (scratch_dir + self.file_names[4], scratch_dir + self.file_names[5], ),
                 "input_B": (scratch_dir + self.file_names[6], scratch_dir + self.file_names[7], ),
            }

            value = self.io.get_local_files("train")
            self.assertDictEqual(target_dirs_train, value)

            value = self.io.get_local_files("val")
            self.assertDictEqual(target_dirs_val, value)

        finally:
            # reset the env variable
            if tempdir_environ is not None:
                os.environ["TMPDIR"] = tempdir_environ
            else:
                os.environ.pop("TMPDIR")

            shutil.rmtree(temp_temp_dir)

    def test_check_connections_no_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        self.io.cfg.label_modifier = get_dummy_label_modifier(output_shapes.keys())
        model = build_dummy_model(input_shapes, output_shapes)

        self.io.check_connections(model)

    def test_check_connections_ok_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        def sample_modifier(info_blob):
            x_values = info_blob["x_values"]
            return {'input_A': x_values["input_A"], 'input_B': x_values["input_B"]}

        self.io.cfg.label_modifier = get_dummy_label_modifier(
            output_shapes.keys())
        self.io.cfg.sample_modifier = sample_modifier

        model = build_dummy_model(input_shapes, output_shapes)
        self.io.check_connections(model)

    def test_check_connections_wrong_sample(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        def sample_modifier(info_blob):
            x_values = info_blob["x_values"]
            return {'input_A': x_values["input_A"]}

        self.io.cfg.label_modifier = get_dummy_label_modifier(
            output_shapes.keys())
        self.io.cfg.sample_modifier = sample_modifier

        model = build_dummy_model(input_shapes, output_shapes)
        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_wrong_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        self.io.cfg.label_modifier = get_dummy_label_modifier(["out_A"])
        model = build_dummy_model(input_shapes, output_shapes)

        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_no_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "out_A": 1,
            "out_B": 1,
        }

        model = build_dummy_model(input_shapes, output_shapes)

        with self.assertRaises(ValueError):
            self.io.check_connections(model)

    def test_check_connections_auto_label(self):
        input_shapes = self.n_bins
        output_shapes = {
            "mc_A": 1,
            "mc_B": 1,
        }

        model = build_dummy_model(input_shapes, output_shapes)

        self.io.check_connections(model)

    def test_get_n_bins(self):
        value = self.io.get_n_bins()
        self.assertSequenceEqual(value, self.n_bins)

    def test_get_input_shape(self):
        value = self.io.get_input_shapes()
        self.assertSequenceEqual(value, self.n_bins)

        def sample_modifier(info_blob):
            x_values = info_blob["x_values"]
            return {'input_A': x_values["input_A"], }
        self.io.cfg.sample_modifier = sample_modifier

        value = self.io.get_input_shapes()
        self.assertEqual(value, {"input_A": self.n_bins["input_A"]})

    def test_get_file_sizes_train(self):
        value = self.io.get_file_sizes("train")
        self.assertSequenceEqual(value, self.train_sizes)

    def test_get_batch_xs(self):
        value = self.io.get_batch()
        target = {
            "input_A": self.train_A_file_1_ctnt[0][:self.batchsize],
            "input_B": self.train_B_file_1_ctnt[0][:self.batchsize],
        }
        assert_dict_arrays_equal(value["x_values"], target)

    def test_get_batch_ys(self):
        value = self.io.get_batch()
        target = {
            "input_A": self.train_A_file_1_ctnt[1][:self.batchsize],
            "input_B": self.train_B_file_1_ctnt[1][:self.batchsize],
        }
        assert_equal_struc_array(value["y_values"], target["input_A"])

    def test_get_all_epochs(self):
        epochs = self.io.get_all_epochs()
        target = [
            (1, 1), (1, 2), (2, 1),
        ]
        self.assertSequenceEqual(epochs, target)

    def test_get_latest_epoch(self):
        value = self.io.get_latest_epoch()
        target = (2, 1)
        self.assertSequenceEqual(value, target)

    def test_get_latest_epoch_no_epochs(self):
        self.io.cfg.output_folder = "./missing/"
        value = self.io.get_latest_epoch()
        self.assertEqual(value, None)

    def test_get_next_epoch_none(self):
        value = self.io.get_next_epoch(None)
        target = (1, 1)
        self.assertSequenceEqual(value, target)

    def test_get_next_epoch_1_1(self):
        value = self.io.get_next_epoch((1, 1))
        target = (1, 2)
        self.assertSequenceEqual(value, target)

    def test_get_next_epoch_1_2(self):
        value = self.io.get_next_epoch((1, 2))
        target = (2, 1)
        self.assertSequenceEqual(value, target)

    def test_get_previous_epoch_2_1(self):
        value = self.io.get_previous_epoch((2, 1))
        target = (1, 2)
        self.assertSequenceEqual(value, target)

    def test_get_previous_epoch_1_2(self):
        value = self.io.get_previous_epoch((1, 2))
        target = (1, 1)
        self.assertSequenceEqual(value, target)

    def test_get_model_path(self):
        value = self.io.get_model_path(1, 1)
        target = self.output_folder + '/saved_models/model_epoch_1_file_1.h5'
        self.assertEqual(value, target)

    def test_get_model_path_local(self):
        value = self.io.get_model_path(1, 1, local=True)
        target = 'saved_models/model_epoch_1_file_1.h5'
        self.assertEqual(value, target)

    def test_get_model_path_latest(self):
        value = self.io.get_model_path(-1, -1)
        target = self.output_folder + '/saved_models/model_epoch_2_file_1.h5'
        self.assertEqual(value, target)

    def test_get_model_path_latest_invalid(self):
        with self.assertRaises(ValueError):
            self.io.get_model_path(1, -1)

        with self.assertRaises(ValueError):
            self.io.get_model_path(-1, 1)

    def test_get_pred_files_list(self):
        value = self.io.get_pred_files_list()
        target = self.pred_filepaths
        self.assertEqual(value, target)

    def test_get_pred_files_list_epoch_given(self):
        value = self.io.get_pred_files_list(epoch=2)
        target = self.pred_filepaths
        self.assertEqual(value, target)

    def test_get_pred_files_list_fileno_given(self):
        value = self.io.get_pred_files_list(fileno=2)
        target = self.pred_filepaths
        self.assertEqual(value, target)

    def test_get_pred_files_list_epoch_fileno_given(self):
        value = self.io.get_pred_files_list(epoch=2, fileno=2)
        target = self.pred_filepaths
        self.assertEqual(value, target)

    def test_get_pred_files_list_no_files(self):
        value = self.io.get_pred_files_list(epoch=3)
        target = []
        self.assertEqual(value, target)

    def test_get_latest_prediction_file_no(self):
        value = self.io.get_latest_prediction_file_no(2, 2)
        target = 2
        self.assertEqual(value, target)

    def test_get_pred_path(self):
        value = self.io.get_pred_path(1, 2, 3)
        target = self.io.get_subfolder("predictions") + '/pred_model_epoch_1_file_2_on_in_out_test_list_val_file_3.h5'
        self.assertEqual(value, target)

    def test_get_local_files_train(self):
        value = self.io.get_local_files("train")
        target = {
            'input_A': ('input_A_train_1.h5', 'input_A_train_2.h5'),
            'input_B': ('input_B_train_1.h5', 'input_B_train_2.h5'),
        }
        self.assertDictEqual(value, target)

    def test_get_local_files_val(self):
        value = self.io.get_local_files("val")
        target = {
            'input_A': ('input_A_val_1.h5', 'input_A_val_2.h5',),
            'input_B': ('input_B_val_1.h5', 'input_B_val_2.h5',)
        }
        self.assertDictEqual(value, target)

    def test_get_no_of_files_train(self):
        value = self.io.get_no_of_files("train")
        target = 2
        self.assertEqual(value, target)

    def test_get_no_of_files_val(self):
        value = self.io.get_no_of_files("val")
        target = 2
        self.assertEqual(value, target)

    def test_yield_files_train(self):
        file_paths = self.io.yield_files("train")
        target = (
            {
                'input_A': 'input_A_train_1.h5',
                'input_B': 'input_B_train_1.h5',
            },
            {
                'input_A': 'input_A_train_2.h5',
                'input_B': 'input_B_train_2.h5',
            },
        )
        for i, value in enumerate(file_paths):
            self.assertDictEqual(value, target[i])

    def test_yield_files_val(self):
        file_paths = self.io.yield_files("val")
        target = (
            {
                'input_A': 'input_A_val_1.h5',
                'input_B': 'input_B_val_1.h5',
            },
            {
                'input_A': 'input_A_val_2.h5',
                'input_B': 'input_B_val_2.h5',
            },
        )
        for i, value in enumerate(file_paths):
            self.assertDictEqual(value, target[i])

    def test_get_file_train(self):
        value = self.io.get_file("train", 2)
        target = {
                'input_A': 'input_A_train_2.h5',
                'input_B': 'input_B_train_2.h5',
        }
        self.assertDictEqual(value, target)

    def test_get_file_val(self):
        value = self.io.get_file("val", 1)
        target = {
                'input_A': 'input_A_val_1.h5',
                'input_B': 'input_B_val_1.h5',
        }
        self.assertDictEqual(value, target)


class TestIOHandlerLR(TestCase):
    """
    Test the learning rate method of the io handler. The csv part
    will generate files in the temp dir.
    """
    def setUp(self):
        self.temp_dir = os.path.join(
            os.path.dirname(__file__), ".temp", "TestIOHandlerLR")
        os.makedirs(self.temp_dir)

        cfg = Configuration(self.temp_dir, None, None)
        self.io = IOHandler(cfg)
        self.get_learning_rate = self.io.get_learning_rate

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_get_learning_rate_float(self):
        user_lr = 0.1
        no_train_files = 3

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        for fileno in range(no_train_files):
            lr = self.get_learning_rate((1, fileno))
            self.assertEqual(lr, user_lr)

        for fileno in range(no_train_files):
            lr = self.get_learning_rate((6, fileno))
            self.assertEqual(lr, user_lr)

    def test_get_learning_rate_tuple(self):
        user_lr = (0.1, 0.2)
        no_train_files = 3

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        rates_epoch_1 = [0.1, 0.08, 0.064]
        for fileno in range(no_train_files):
            lr = self.get_learning_rate((1, fileno+1))
            self.assertAlmostEqual(lr, rates_epoch_1[fileno])

        rates_epoch_2 = [0.0512, 0.04096, 0.032768]
        for fileno in range(no_train_files):
            lr = self.get_learning_rate((2, fileno + 1))
            self.assertAlmostEqual(lr, rates_epoch_2[fileno])

    def test_get_learning_rate_function(self):
        def get_lr(epoch, filenos):
            return epoch+filenos

        user_lr = get_lr
        no_train_files = 3

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        rates_epoch_1 = [2, 3, 4]
        for fileno in range(no_train_files):
            lr = self.get_learning_rate((1, fileno+1))
            self.assertAlmostEqual(lr, rates_epoch_1[fileno])

        rates_epoch_2 = [3, 4, 5]
        for fileno in range(no_train_files):
            lr = self.get_learning_rate((2, fileno + 1))
            self.assertAlmostEqual(lr, rates_epoch_2[fileno])

    def test_get_learning_rate_other(self):
        user_lr = print
        no_train_files = 3

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        with self.assertRaises(TypeError):
            self.get_learning_rate((1, 1))

    def test_get_learning_rate_read_csv_file(self):
        user_lr = "lr.csv"
        no_train_files = "not used"

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        csv_content = (
            "1  2   0.1\n"
            "1  3   0.2\n"
            "2  2   0.3\n"
        )

        with open(os.path.join(self.temp_dir, "lr.csv"), "w") as f:
            f.write(csv_content)

        with self.assertRaises(ValueError):
            self.get_learning_rate((1, 1))

        tests = [
            {"inp": (1, 2), "target": 0.1},
            {"inp": (1, 3), "target": 0.2},
            {"inp": [1, 10], "target": 0.2},
            {"inp": (2, 1), "target": 0.2},
            {"inp": (2, 2), "target": 0.3},
            {"inp": (2, 3), "target": 0.3},
            {"inp": (25, 33), "target": 0.3},
        ]
        for test in tests:
            self.assertEqual(
                self.get_learning_rate(test["inp"]), test["target"])

    def test_get_learning_rate_read_csv_file_only_one_line(self):
        user_lr = "lr.csv"
        no_train_files = "not used"

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        csv_content = (
            "0  1   2\n"
        )

        with open(os.path.join(self.temp_dir, "lr.csv"), "w") as f:
            f.write(csv_content)

        tests = [
            {"inp": (0, 1), "target": 2},
        ]
        for test in tests:
            self.assertEqual(
                self.get_learning_rate(test["inp"]), test["target"])

    def test_get_learning_rate_read_csv_file_bad_format_too_short(self):
        user_lr = "lr.csv"
        no_train_files = "not used"

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        csv_content = (
            "0  1\n"
            "0  1\n"
        )

        with open(os.path.join(self.temp_dir, "lr.csv"), "w") as f:
            f.write(csv_content)

        with self.assertRaises(ValueError):
            self.get_learning_rate((1, 1))

    def test_get_learning_rate_read_csv_file_bad_format_too_long(self):
        user_lr = "lr.csv"
        no_train_files = "not used"

        self.io.cfg.learning_rate = user_lr
        self.io.get_no_of_files = MagicMock(return_value=no_train_files)

        csv_content = (
            "0  1   2   4\n"
            "0  1   2   3\n"
        )

        with open(os.path.join(self.temp_dir, "lr.csv"), "w") as f:
            f.write(csv_content)

        with self.assertRaises(ValueError):
            self.get_learning_rate((1, 1))


class TestFunctions(TestCase):
    def test_split_name_of_predfile(self):
        filename = "pred_model_epoch_1_file_2_on_list_val_file_3.h5"
        target = (1, 2, 3)

        self.assertSequenceEqual(split_name_of_predfile(filename), target)


def assert_equal_struc_array(a, b):
    """  np.testing.assert_array_equal does not work for arrays containing nans...
     so test individual instead. """
    np.testing.assert_array_equal(a.dtype, b.dtype)
    for name in a.dtype.names:
        np.testing.assert_almost_equal(a[name], b[name])


def save_dummy_h5py(path, shape, value_xs=1., value_ys=1., size=50):
    xs = np.ones((size,) + shape) * value_xs

    dtypes = [('mc_A', '<f8'), ('mc_B', '<f8'), ]
    ys = (np.ones((size, 2))*value_ys).ravel().view(dtype=dtypes)

    with h5py.File(path, 'w') as h5f:
        h5f.create_dataset('x', data=xs, dtype='<f8')
        h5f.create_dataset('y', data=ys, dtype=dtypes)

    return xs, ys


def build_dummy_model(input_shapes, output_shapes):
    """

    Parameters
    ----------
    input_shapes : dict
    output_shapes : dict

    Returns
    -------
    model : keras model

    """
    inputs = {}
    for name, shape in input_shapes.items():
        inputs[name] = layers.Input(shape, name=name)
    conc = layers.Concatenate()(list(inputs.values()))
    flat = layers.Flatten()(conc)

    outputs = {}
    for name, shape in output_shapes.items():
        outputs[name] = layers.Dense(shape, name=name)(flat)

    model = Model(list(inputs.values()), list(outputs.values()))
    return model


def get_dummy_label_modifier(output_names):
    def label_modifier(info_blob):
        y_values = info_blob["y_values"]
        particle = y_values["mc_A"]
        y_true = dict()
        for output_name in output_names:
            y_true[output_name] = particle
        return y_true
    return label_modifier


def assert_dict_arrays_equal(a, b):
    for key, value in a.items():
        assert key in b
        np.testing.assert_array_almost_equal(a[key], b[key])




import numpy as np
import orcanet.misc as misc
from unittest import TestCase


class TestFunctions(TestCase):
    def test_dict_to_recarray_shape(self):
        inp = {"aa": np.ones((5, 3)), "bb": np.ones((5, 1))}
        output = misc.dict_to_recarray(inp)
        self.assertTrue(output.shape == (5, ))

    def test_dict_to_recarray_dtype(self):
        inp = {"aa": np.ones((5, 3), dtype="int16"), "bb": np.ones((5, 1), dtype="float32")*4}
        output = misc.dict_to_recarray(inp)
        dtypes = {
            "aa_1": np.dtype("int16"),
            "aa_2": np.dtype("int16"),
            "aa_3": np.dtype("int16"),
            "bb_1": np.dtype("float32"),
        }
        self.assertTupleEqual(output.dtype.names, ('aa_1', 'aa_2', 'aa_3', 'bb_1'))
        for dtype_name, dtype in dtypes.items():
            np.testing.assert_equal(output[dtype_name].dtype, dtype)
        np.testing.assert_equal(output["aa_1"], np.ones(5, dtype="int16"))
        np.testing.assert_equal(output["bb_1"], np.ones(5, dtype="float32")*4)

    def test_dict_to_recarray_len_1(self):
        inp = {"aa": np.ones((5, 3)), "bb": np.ones((5, ))}
        output = misc.dict_to_recarray(inp)
        self.assertTrue(output.shape == (5, ))
        self.assertTupleEqual(output.dtype.names, ('aa_1', 'aa_2', 'aa_3', 'bb_1'))

    def test_dict_to_recarray_wrong_dim(self):
        inp = {"aa": np.ones((4, 3)), "bb": np.ones((5, 1))}
        with self.assertRaises(ValueError):
            misc.dict_to_recarray(inp)

    def test_dict_to_recarray_weird_shape(self):
        inp = {"aa": np.ones((5, 3, 2)), "bb": np.ones((5, 1))}
        output = misc.dict_to_recarray(inp)
        self.assertTrue(output.shape == (5, ))
        self.assertTupleEqual(
            output.dtype.names,
            ('aa_1', 'aa_2', 'aa_3', 'aa_4', 'aa_5', 'aa_6', 'bb_1'))


class TestRegister(TestCase):
    def setUp(self):
        self.register = {
            "_func": _func,
            "_Cls": _Cls,
        }

    def _parse_entry(self, toml_entry):
        return misc.from_register(toml_entry, self.register)

    def test_func_str(self):
        self.assertEqual(self._parse_entry("_func")(12), 12)

    def test_func_list(self):
        with self.assertRaises(TypeError):
            self._parse_entry(["_func", 5])

    def test_class_str(self):
        with self.assertRaises(TypeError):
            self._parse_entry("_Cls")

    def test_class_list(self):
        self.assertTupleEqual(self._parse_entry(["_Cls", 0])(12), (0, 12))

    def test_class_dict(self):
        self.assertTupleEqual(
            self._parse_entry({"name": "_Cls", "a": 0})(12), (0, 12))

    def test_class_dict_name_missing(self):
        with self.assertRaises(KeyError):
            self._parse_entry({"names": "_Cls", "a": 0})(12), (0, 12)

    def test_class_list_dict(self):
        self.assertTupleEqual(
            self._parse_entry(["_Cls", {"a": 0}])(12), (0, 12))

    def test_register(self):
        saved, register = misc.get_register()
        register(_func)
        register(_Cls)
        self.assertDictEqual(self.register, saved)


def _func(a):
    return a


class _Cls:
    def __init__(self, a):
        self.a = a

    def __call__(self, x):
        return self.a, x
